/*
 * HyperV  Detection code.
 *
 * Copyright (C) 2010, Novell, Inc.
 * Author : K. Y. Srinivasan <ksrinivasan@novell.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, GOOD TITLE or
 * NON INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <linux/init.h>
#include <linux/types.h>
#include <linux/clocksource.h>
#include <linux/dmi.h>
#include <linux/err.h>
#include <asm/processor.h>
#include <asm/hyperv.h>
#include <asm/mshyperv.h>
#include <asm/div64.h>

#define HV_CLOCK_SHIFT	22

unsigned int use_hv_time = 0;

int ms_hyperv_platform(void)
{
	u32 eax, ebx, ecx, edx;
	char hyp_signature[13];

	cpuid(1, &eax, &ebx, &ecx, &edx);
	if (!(ecx & HYPERV_HYPERVISOR_PRESENT_BIT))
		return 0;

	cpuid(HYPERV_CPUID_VENDOR_AND_MAX_FUNCTIONS, &eax, &ebx, &ecx, &edx);
	*(u32 *)(hyp_signature + 0) = ebx;
	*(u32 *)(hyp_signature + 4) = ecx;
	*(u32 *)(hyp_signature + 8) = edx;

	if ((eax < HYPERV_CPUID_MIN) || (memcmp("Microsoft Hv", hyp_signature, 12)))
		return 0;
	return 1;
}

cycle_t read_hv_clock(void)
{
	cycle_t current_tick;
	/*
	 * Read the partition counter to get the current tick count. This count
	 * is set to 0 when the partition is created and is incremented in
	 * 100 nanosecond units.
	 */
	rdmsrl(HV_X64_MSR_TIME_REF_COUNT, current_tick);
	return current_tick;
}

void __cpuinit ms_hyperv_set_feature_bits(struct cpuinfo_x86 *c)
{
	u32 eax, ebx, ecx, edx;

	c->x86_hyper_features = 0;
	/*
	 * Extract the features, recommendations etc.
	 * The first 9 bits will be used to track hypervisor features.
	 * The next 6 bits will be used to track the hypervisor
	 * recommendations.
	 */
	cpuid(HYPERV_CPUID_FEATURES, &eax, &ebx, &ecx, &edx);
	c->x86_hyper_features |= (eax & 0x1ff);

	cpuid(HYPERV_CPUID_ENLIGHTMENT_INFO, &eax, &ebx, &ecx, &edx);
	c->x86_hyper_features |= ((eax & 0x3f) << 9);
	printk(KERN_INFO "Detected HyperV with features: %x\n",
		c->x86_hyper_features);
}

#ifdef CONFIG_X86_32
static struct clocksource hyperv_cs = {
	.name           = "hyperv_clocksource",
	.rating         = 499, /* use this when running on Hyperv*/
	.read           = read_hv_clock,
	.mask           = CLOCKSOURCE_MASK(64),
	/*
	 * The time ref counter in HyperV is in 100ns units.
	 * The definition of mult is:
	 * mult/2^shift = ns/cyc = 100
	 * mult = (100 << shift)
	 */
	.mult           = (100 << HV_CLOCK_SHIFT),
	.shift          = HV_CLOCK_SHIFT,
};
#endif

void __cpuinit init_hv_clocksource(void)
{
	if (!(boot_cpu_data.x86_hyper_vendor == X86_HYPER_VENDOR_MSFT) ||
	    !(boot_cpu_data.x86_hyper_features &
		    HV_X64_MSR_TIME_REF_COUNT_AVAILABLE))
		return;

	use_hv_time = 1;
	pr_info("Registering HyperV clock source\n");
#ifdef CONFIG_X86_32
	clocksource_register(&hyperv_cs);
#endif
}

extern  u32 pmtmr_ioport;
#define PMTMR_TICKS_PER_SEC 3579545
#define ACPI_PM_MASK ((1ULL<<(24))-1) /* Limit it to 24 bits */

static inline u32 acpi_pm_read(void)
{
	u32 v1 = 0, v2 = 0, v3 = 0;
	if (!pmtmr_ioport)
		return 0;
	do {
		v1 = inl(pmtmr_ioport);
		v2 = inl(pmtmr_ioport);
		v3 = inl(pmtmr_ioport);
	} while (unlikely((v1 > v2 && v1 < v3) || (v2 > v3 && v2 < v1)
			  || (v3 > v1 && v3 < v2)));

	return (v2 & ACPI_PM_MASK);
}

static unsigned long __hyperv_get_tsc_khz(void)
{
	u64 tsc_start, tsc_end;
	u64 pm_start, pm_end;
	u64 start_count, end_count, delta, pm_delta;
	unsigned long frequency_hv, frequency_pm;

	/*
	 * The HyperV partition time reference counter is 100ns granular
	 * monotonically increasing counter. Calibrate TSC against this
	 * time source over a period of  50 ms.
	 */
again:
	tsc_start = get_cycles_sync();
	pm_start = acpi_pm_read();
	rdmsrl(HV_X64_MSR_TIME_REF_COUNT, start_count);
	do {
		rdmsrl(HV_X64_MSR_TIME_REF_COUNT, end_count);
		delta = end_count - start_count;
	} while (delta < (10*1000*50));
	tsc_end = get_cycles_sync();
	pm_end = acpi_pm_read();
	if (tsc_start > tsc_end)
		goto again;
	if (pm_end < pm_start)
		pm_end += (u64)(1<<24);
	pm_delta = (pm_end - pm_start);
	do_div(delta, 10);	/* delta in useconds */
	frequency_hv = ((tsc_end - tsc_start) * 1000); /* in KHZ */
	do_div(frequency_hv, delta);
	if (pm_delta) {
		pm_delta *= 1000000000LL;
		do_div(pm_delta, PMTMR_TICKS_PER_SEC);  /* in nano seconds */
		tsc_end -= tsc_start;
		frequency_pm = tsc_end * 1000000L;
		do_div(frequency_pm, pm_delta); /* in KHZ */
		frequency_hv = min(frequency_hv, frequency_pm);
	}
	return frequency_hv;
}

unsigned long hyperv_get_tsc_khz(void)
{
	int i;
	unsigned long tsc;
	unsigned long tsc_min = ULONG_MAX;

	if (use_hv_time > 0) {
		/* Calibrate 5 times and pick the lowest.  */
		for (i=0; i < 5; i++) {
			tsc = __hyperv_get_tsc_khz();
			if (tsc_min > tsc)
				tsc_min = tsc;
		}
		return tsc_min;
	} else
		return 0; /* indicates we should not use this value */
}
