/*
 * Copyright (c) 2005-2012 Brocade Communications Systems, Inc.
 * All rights reserved
 * www.brocade.com
 *
 * Linux driver for Brocade Fibre Channel Host Bus Adapter.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License (GPL) Version 2 as
 * published by the Free Software Foundation
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <linux/uaccess.h>

#include "bfad_drv.h"
#include "bfad_im.h"
#include "bfad_ioctl.h"

BFA_TRC_FILE(LDRV, IOCTL);

static bfa_boolean_t
bfad_is_ready(void)
{
	struct bfad_s *bfad = NULL;

	mutex_lock(&bfad_mutex);
	if (!bfad_scan_done) {
		mutex_unlock(&bfad_mutex);
		return BFA_FALSE;
	}

	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (!(bfad->bfad_flags & BFAD_DRV_INIT_DONE)) {
			mutex_unlock(&bfad_mutex);
			return BFA_FALSE;
		}
	}
	mutex_unlock(&bfad_mutex);
	return BFA_TRUE;
}

struct bfad_s *
bfad_find_bfad_by_inst_no(int inst_no)
{
	struct bfad_s *bfad = NULL, *rc = NULL;

	mutex_lock(&bfad_mutex);
	list_for_each_entry(bfad, &bfad_list, list_entry) {
		if (bfad->inst_no == inst_no) {
			rc = bfad;
			break;
		}
	}
	mutex_unlock(&bfad_mutex);
	return rc;
}

int
bfad_iocmd_ioc_enable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	int	rc = 0;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	/* If IOC is not in disabled state - return */
	if (!bfa_ioc_is_disabled(&bfad->bfa.ioc)) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_IOC_FAILURE;
		return rc;
	}

	init_completion(&bfad->enable_comp);
	bfa_iocfc_enable(&bfad->bfa);
	iocmd->status = BFA_STATUS_OK;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	wait_for_completion(&bfad->enable_comp);

	return rc;
}

int
bfad_iocmd_ioc_disable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	int	rc = 0;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfad->disable_active) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return -EBUSY;
	}

	bfad->disable_active = BFA_TRUE;
	init_completion(&bfad->disable_comp);
	bfa_iocfc_disable(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	wait_for_completion(&bfad->disable_comp);
	bfad->disable_active = BFA_FALSE;
	iocmd->status = BFA_STATUS_OK;

	return rc;
}

static int
bfad_iocmd_ioc_get_info(void *cmd)
{
	int	i;
	struct bfa_ioctl_ioc_info_s *iocmd = (struct bfa_ioctl_ioc_info_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_im_port_s	*im_port;
	struct bfa_port_attr_s	pattr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcport_get_attr(&bfad->bfa, &pattr);
	iocmd->nwwn = pattr.nwwn;
	iocmd->pwwn = pattr.pwwn;
	iocmd->ioc_type = bfa_get_type(&bfad->bfa);
	iocmd->mac = bfa_get_mac(&bfad->bfa);
	iocmd->factory_mac = bfa_get_mfg_mac(&bfad->bfa);
	bfa_get_adapter_serial_num(&bfad->bfa, iocmd->serialnum);
	iocmd->factorynwwn = pattr.factorynwwn;
	iocmd->factorypwwn = pattr.factorypwwn;
	iocmd->bfad_num = bfad->inst_no;
	im_port = bfad->pport.im_port;
	iocmd->host = im_port->shost->host_no;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	strcpy(iocmd->name, bfad->adapter_name);
	strcpy(iocmd->port_name, bfad->port_name);
	strcpy(iocmd->hwpath, bfad->pci_name);

	/* set adapter hw path */
	strcpy(iocmd->adapter_hwpath, bfad->pci_name);
	i = strlen(iocmd->adapter_hwpath) - 1;
	while (iocmd->adapter_hwpath[i] != '.')
		i--;
	iocmd->adapter_hwpath[i] = '\0';
	iocmd->status = BFA_STATUS_OK;
	return 0;
}

static int
bfad_iocmd_ioc_get_attr(void *cmd)
{
	struct bfa_ioctl_ioc_attr_s *iocmd = (struct bfa_ioctl_ioc_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_ioc_get_attr(&bfad->bfa.ioc, &iocmd->ioc_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	/* fill in driver attr info */
	strcpy(iocmd->ioc_attr.driver_attr.driver, BFAD_DRIVER_NAME);
	strncpy(iocmd->ioc_attr.driver_attr.driver_ver,
		BFAD_DRIVER_VERSION, BFA_VERSION_LEN);
	strcpy(iocmd->ioc_attr.driver_attr.fw_ver,
		iocmd->ioc_attr.adapter_attr.fw_ver);
	strcpy(iocmd->ioc_attr.driver_attr.bios_ver,
		iocmd->ioc_attr.adapter_attr.optrom_ver);

	/* copy chip rev info first otherwise it will be overwritten */
	memcpy(bfad->pci_attr.chip_rev, iocmd->ioc_attr.pci_attr.chip_rev,
		sizeof(bfad->pci_attr.chip_rev));
	memcpy(&iocmd->ioc_attr.pci_attr, &bfad->pci_attr,
		sizeof(struct bfa_ioc_pci_attr_s));

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_ioc_get_stats(void *cmd)
{
	struct bfa_ioctl_ioc_stats_s *iocmd =
				(struct bfa_ioctl_ioc_stats_s *)cmd;
	struct bfad_s *bfad = NULL;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	bfa_ioc_get_stats(&bfad->bfa, &iocmd->ioc_stats);
	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_ioc_get_fwstats(void *cmd)
{
	struct bfa_ioctl_ioc_fwstats_s *iocmd =
			(struct bfa_ioctl_ioc_fwstats_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->buf_size);
	if (!iocmd_bufptr) {
		bfa_trc(bfad, 0x4444);
		return -ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ioc_fw_stats_get(&bfad->bfa.ioc, iocmd_bufptr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->buf_size)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	bfa_trc(bfad, 0x6666);
	return 0;
}

int
bfad_iocmd_ioc_reset_stats(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (v_cmd == IOCMD_IOC_RESET_STATS) {
		bfa_ioc_clear_stats(&bfad->bfa);
		iocmd->status = BFA_STATUS_OK;
	} else if (v_cmd == IOCMD_IOC_RESET_FWSTATS) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		iocmd->status = bfa_ioc_fw_stats_clear(&bfad->bfa.ioc);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	}

	return 0;
}

int
bfad_iocmd_ioc_set_name(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_ioc_name_s *iocmd = (struct bfa_ioctl_ioc_name_s *)cmd;
	struct bfad_s *bfad = NULL;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (v_cmd == IOCMD_IOC_SET_ADAPTER_NAME)
		strcpy(bfad->adapter_name, iocmd->name);
	else if (v_cmd == IOCMD_IOC_SET_PORT_NAME)
		strcpy(bfad->port_name, iocmd->name);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_iocfc_get_attr(void *cmd)
{
	struct bfa_ioctl_iocfc_attr_s *iocmd =
				(struct bfa_ioctl_iocfc_attr_s *)cmd;
	struct bfad_s *bfad = NULL;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd->status = BFA_STATUS_OK;
	bfa_iocfc_get_attr(&bfad->bfa, &iocmd->iocfc_attr);

	return 0;
}

int
bfad_iocmd_iocfc_set_intr(void *cmd)
{
	struct bfa_ioctl_iocfc_intr_s *iocmd =
				(struct bfa_ioctl_iocfc_intr_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_iocfc_israttr_set(&bfad->bfa, &iocmd->attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_port_enable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_port_enable(&bfad->bfa.modules.port,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		return 0;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	return 0;
}

int
bfad_iocmd_port_disable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_port_disable(&bfad->bfa.modules.port,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		return 0;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	return 0;
}

static int
bfad_iocmd_port_get_attr(void *cmd)
{
	struct bfa_ioctl_port_attr_s *iocmd =
				(struct bfa_ioctl_port_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_lport_attr_s	port_attr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcport_get_attr(&bfad->bfa, &iocmd->attr);
	bfa_fcs_lport_get_attr(&bfad->bfa_fcs.fabric.bport, &port_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->attr.topology != BFA_PORT_TOPOLOGY_NONE)
		iocmd->attr.pid = port_attr.pid;
	else
		iocmd->attr.pid = 0;

	iocmd->attr.port_type = port_attr.port_type;
	iocmd->attr.loopback = port_attr.loopback;
	iocmd->attr.authfail = port_attr.authfail;
	strncpy(iocmd->attr.port_symname.symname,
		port_attr.port_cfg.sym_name.symname,
		sizeof(port_attr.port_cfg.sym_name.symname));

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_port_get_stats(void *cmd)
{
	struct bfa_ioctl_port_stats_s *iocmd =
				(struct bfa_ioctl_port_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	void	*iocmd_bufptr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->buf_size);
	if (!iocmd_bufptr)
		return -ENOMEM;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_port_get_stats(&bfad->bfa.modules.port,
				iocmd_bufptr, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->buf_size)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_port_reset_stats(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_port_clear_stats(&bfad->bfa.modules.port,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		return 0;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	return 0;
}

int
bfad_iocmd_set_port_cfg(void *iocmd, unsigned int v_cmd)
{
	struct bfa_ioctl_port_cfg_s *cmd = (struct bfa_ioctl_port_cfg_s *)iocmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(cmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (v_cmd == IOCMD_PORT_CFG_TOPO)
		cmd->status = bfa_fcport_cfg_topology(&bfad->bfa, cmd->param);
	else if (v_cmd == IOCMD_PORT_CFG_SPEED)
		cmd->status = bfa_fcport_cfg_speed(&bfad->bfa, cmd->param);
	else if (v_cmd == IOCMD_PORT_CFG_ALPA)
		cmd->status = bfa_fcport_cfg_hardalpa(&bfad->bfa, cmd->param);
	else if (v_cmd == IOCMD_PORT_CLR_ALPA)
		cmd->status = bfa_fcport_clr_hardalpa(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_port_cfg_maxfrsize(void *cmd)
{
	struct bfa_ioctl_port_cfg_maxfrsize_s *iocmd =
				(struct bfa_ioctl_port_cfg_maxfrsize_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcport_cfg_maxfrsize(&bfad->bfa, iocmd->maxfrsize);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_port_cfg_bbsc(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfa_ioc_get_type(&bfad->bfa.ioc) == BFA_IOC_TYPE_FC) {
		if (v_cmd == IOCMD_PORT_BBSC_ENABLE)
			fcport->cfg.bb_scn_state = BFA_TRUE;
		else if (v_cmd == IOCMD_PORT_BBSC_DISABLE)
			fcport->cfg.bb_scn_state = BFA_FALSE;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

static int
bfad_iocmd_lport_get_attr(void *cmd)
{
	struct bfa_fcs_lport_s	*fcs_port;
	struct bfa_ioctl_lport_attr_s *iocmd =
				(struct bfa_ioctl_lport_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_lport_get_attr(fcs_port, &iocmd->port_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_lport_get_stats(void *cmd)
{
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_ioctl_lport_stats_s *iocmd =
			(struct bfa_ioctl_lport_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_lport_get_stats(fcs_port, &iocmd->port_stats);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_lport_reset_stats(void *cmd)
{
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_ioctl_reset_stats_s *iocmd =
			(struct bfa_ioctl_reset_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcpim_s *fcpim;
	struct list_head *qe, *qen;
	struct bfa_itnim_s *itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcpim = BFA_FCPIM(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->vpwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_lport_clear_stats(fcs_port);
	/* clear IO stats from all active itnims */
	list_for_each_safe(qe, qen, &fcpim->itnim_q) {
		itnim = (struct bfa_itnim_s *) qe;
		if (itnim->rport->rport_info.lp_tag != fcs_port->lp_tag)
			continue;
		bfa_itnim_clear_stats(itnim);
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_lport_get_iostats(void *cmd)
{
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_ioctl_lport_iostats_s *iocmd =
			(struct bfa_ioctl_lport_iostats_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcpim_port_iostats(&bfad->bfa, &iocmd->iostats,
			fcs_port->lp_tag);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_lport_get_rports(void *cmd)
{
	struct bfa_ioctl_lport_get_rports_s *iocmd =
			(struct bfa_ioctl_lport_get_rports_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	unsigned long	flags;
	void	*iocmd_bufptr;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (iocmd->nrports == 0)
		return -EINVAL;

	iocmd_bufptr = vzalloc(sizeof(wwn_t) * iocmd->nrports);
	if (iocmd_bufptr == NULL) {
		bfa_trc(bfad, 0);
		iocmd->status = BFA_STATUS_ENOMEM;
		goto out;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		bfa_trc(bfad, 0);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	bfa_fcs_lport_get_rports(fcs_port, (wwn_t *)iocmd_bufptr,
				&iocmd->nrports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (copy_to_user((void __user *)(size_t)iocmd->rbuf_ptr,
			 iocmd_bufptr, iocmd->nrports * sizeof(wwn_t))) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}

	iocmd->status = BFA_STATUS_OK;
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_rport_get_attr(void *cmd)
{
	struct bfa_ioctl_rport_attr_s *iocmd =
				(struct bfa_ioctl_rport_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd->rpwwn);
	if (fcs_rport == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	bfa_fcs_rport_get_attr(fcs_rport, &iocmd->attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

static int
bfad_iocmd_rport_get_addr(void *cmd)
{
	struct bfa_ioctl_rport_scsi_addr_s *iocmd =
			(struct bfa_ioctl_rport_scsi_addr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s	*fcs_port;
	struct bfa_fcs_itnim_s	*fcs_itnim;
	struct bfad_itnim_s	*drv_itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd->rpwwn);
	if (fcs_itnim == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	drv_itnim = fcs_itnim->itnim_drv;

	if (drv_itnim && drv_itnim->im_port)
		iocmd->host = drv_itnim->im_port->shost->host_no;
	else {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	iocmd->target = drv_itnim->scsi_tgt_id;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->bus = 0;
	iocmd->lun = 0;
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_rport_get_stats(void *cmd)
{
	struct bfa_ioctl_rport_stats_s *iocmd =
			(struct bfa_ioctl_rport_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd->rpwwn);
	if (fcs_rport == NULL) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	memcpy((void *)&iocmd->stats, (void *)&fcs_rport->stats,
		sizeof(struct bfa_rport_stats_s));
	memcpy((void *)&iocmd->stats.hal_stats,
	       (void *)&(bfa_fcs_rport_get_halrport(fcs_rport)->stats),
	       sizeof(struct bfa_rport_hal_stats_s));

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_rport_clr_stats(void *cmd)
{
	struct bfa_ioctl_rport_reset_stats_s *iocmd =
				(struct bfa_ioctl_rport_reset_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	struct bfa_rport_s *rport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd->rpwwn);
	if (fcs_rport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	memset((char *)&fcs_rport->stats, 0, sizeof(struct bfa_rport_stats_s));
	rport = bfa_fcs_rport_get_halrport(fcs_rport);
	memset(&rport->stats, 0, sizeof(rport->stats));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_rport_set_speed(void *cmd)
{
	struct bfa_ioctl_rport_set_speed_s *iocmd =
				(struct bfa_ioctl_rport_set_speed_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (fcs_port == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		goto out;
	}

	fcs_rport = bfa_fcs_rport_lookup(fcs_port, iocmd->rpwwn);
	if (fcs_rport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		goto out;
	}

	fcs_rport->rpf.assigned_speed  = iocmd->speed;
	/* Set this speed in f/w only if the RPSC speed is not available */
	if (fcs_rport->rpf.rpsc_speed == BFA_PORT_SPEED_UNKNOWN)
		bfa_rport_speed(fcs_rport->bfa_rport, iocmd->speed);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_vport_get_attr(void *cmd)
{
	struct bfa_fcs_vport_s *fcs_vport;
	struct bfa_ioctl_vport_attr_s *iocmd =
				(struct bfa_ioctl_vport_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	bfa_fcs_vport_get_attr(fcs_vport, &iocmd->vport_attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_vport_get_stats(void *cmd)
{
	struct bfa_fcs_vport_s *fcs_vport;
	struct bfa_ioctl_vport_stats_s *iocmd =
				(struct bfa_ioctl_vport_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	memcpy((void *)&iocmd->vport_stats, (void *)&fcs_vport->vport_stats,
		sizeof(struct bfa_vport_stats_s));
	memcpy((void *)&iocmd->vport_stats.port_stats,
	       (void *)&fcs_vport->lport.stats,
		sizeof(struct bfa_lport_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_vport_clr_stats(void *cmd)
{
	struct bfa_fcs_vport_s *fcs_vport;
	struct bfa_ioctl_reset_stats_s *iocmd =
				(struct bfa_ioctl_reset_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vport = bfa_fcs_vport_lookup(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->vpwwn);
	if (fcs_vport == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VWWN;
		goto out;
	}

	memset(&fcs_vport->vport_stats, 0, sizeof(struct bfa_vport_stats_s));
	memset(&fcs_vport->lport.stats, 0, sizeof(struct bfa_lport_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

static int
bfad_iocmd_fabric_get_lports(void *cmd)
{
	struct bfa_ioctl_fabric_get_lports_s *iocmd =
			(struct bfa_ioctl_fabric_get_lports_s *)cmd;
	struct bfad_s *bfad = NULL;
	bfa_fcs_vf_t	*fcs_vf;
	u32	nports = iocmd->nports;
	unsigned long	flags;
	void	*iocmd_bufptr;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (nports == 0) {
		iocmd->status = BFA_STATUS_EINVAL;
		return 0;
	}

	iocmd_bufptr = vzalloc(sizeof(wwn_t[iocmd->nports]));
	if (iocmd_bufptr == NULL) {
		bfa_trc(bfad, 0);
		return -ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd->vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}
	bfa_fcs_vf_get_ports(fcs_vf, (wwn_t *)iocmd_bufptr, &nports);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->nports = nports;

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, nports * sizeof(wwn_t))) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}

	iocmd->status = BFA_STATUS_OK;
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_ratelim(unsigned int cmd, void *pcmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)pcmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);

	if (cmd == IOCMD_RATELIM_ENABLE)
		fcport->cfg.ratelimit = BFA_TRUE;
	else if (cmd == IOCMD_RATELIM_DISABLE)
		fcport->cfg.ratelimit = BFA_FALSE;

	if (fcport->cfg.trl_def_speed == BFA_PORT_SPEED_UNKNOWN)
		fcport->cfg.trl_def_speed = BFA_PORT_SPEED_1GBPS;

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;

	return 0;
}

int
bfad_iocmd_ratelim_speed(unsigned int cmd, void *pcmd)
{
	struct bfa_ioctl_trl_speed_s *iocmd =
				(struct bfa_ioctl_trl_speed_s *)pcmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);

	/* Auto and speeds greater than the supported speed, are invalid */
	if ((iocmd->speed == BFA_PORT_SPEED_AUTO) ||
	    (iocmd->speed > fcport->speed_sup)) {
		iocmd->status = BFA_STATUS_UNSUPP_SPEED;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return 0;
	}

	fcport->cfg.trl_def_speed = iocmd->speed;
	iocmd->status = BFA_STATUS_OK;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_cfg_fcpim(void *cmd)
{
	struct bfa_ioctl_fcpim_s *iocmd = (struct bfa_ioctl_fcpim_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	bfa_fcpim_path_tov_set(&bfad->bfa, iocmd->param);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_fcpim_get_modstats(void *cmd)
{
	struct bfa_ioctl_fcpim_modstats_s *iocmd =
			(struct bfa_ioctl_fcpim_modstats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcpim_s *fcpim;
	struct list_head *qe, *qen;
	struct bfa_itnim_s *itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcpim = BFA_FCPIM(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	/* accumulate IO stats from itnim */
	memset((void *)&iocmd->modstats, 0, sizeof(struct bfa_itnim_iostats_s));
	list_for_each_safe(qe, qen, &fcpim->itnim_q) {
		itnim = (struct bfa_itnim_s *) qe;
		bfa_fcpim_add_stats(&iocmd->modstats, &(itnim->stats));
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_fcpim_clr_modstats(void *cmd)
{
	struct bfa_ioctl_fcpim_modstatsclr_s *iocmd =
				(struct bfa_ioctl_fcpim_modstatsclr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcpim_s *fcpim;
	struct list_head *qe, *qen;
	struct bfa_itnim_s *itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcpim = BFA_FCPIM(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	list_for_each_safe(qe, qen, &fcpim->itnim_q) {
		itnim = (struct bfa_itnim_s *) qe;
		bfa_itnim_clear_stats(itnim);
	}
	memset(&fcpim->del_itn_stats, 0,
		sizeof(struct bfa_fcpim_del_itn_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_fcpim_get_del_itn_stats(void *cmd)
{
	struct bfa_ioctl_fcpim_del_itn_stats_s *iocmd =
			(struct bfa_ioctl_fcpim_del_itn_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcpim_s *fcpim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcpim = BFA_FCPIM(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memcpy((void *)&iocmd->modstats, (void *)&fcpim->del_itn_stats,
		sizeof(struct bfa_fcpim_del_itn_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

static int
bfad_iocmd_itnim_get_attr(void *cmd)
{
	struct bfa_ioctl_itnim_attr_s *iocmd =
				(struct bfa_ioctl_itnim_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s	*fcs_port;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->lpwwn);
	if (!fcs_port)
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
	else
		iocmd->status = bfa_fcs_itnim_attr_get(fcs_port,
					iocmd->rpwwn, &iocmd->attr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

static int
bfad_iocmd_itnim_get_iostats(void *cmd)
{
	struct bfa_ioctl_itnim_iostats_s *iocmd =
			(struct bfa_ioctl_itnim_iostats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_itnim_s *itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->lpwwn);
	if (!fcs_port) {
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
	} else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd->rpwwn);
		if (itnim == NULL)
			iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		else {
			iocmd->status = BFA_STATUS_OK;
			memcpy((void *)&iocmd->iostats, (void *)
			       &(bfa_fcs_itnim_get_halitn(itnim)->stats),
			       sizeof(struct bfa_itnim_iostats_s));
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

static int
bfad_iocmd_itnim_reset_stats(void *cmd)
{
	struct bfa_ioctl_rport_reset_stats_s *iocmd =
			(struct bfa_ioctl_rport_reset_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s	*fcs_port;
	struct bfa_fcs_itnim_s	*itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->pwwn);
	if (!fcs_port)
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
	else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd->rpwwn);
		if (itnim == NULL)
			iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		else {
			iocmd->status = BFA_STATUS_OK;
			bfa_fcs_itnim_stats_clear(fcs_port, iocmd->rpwwn);
			bfa_itnim_clear_stats(bfa_fcs_itnim_get_halitn(itnim));
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

static int
bfad_iocmd_itnim_get_itnstats(void *cmd)
{
	struct bfa_ioctl_itnim_itnstats_s *iocmd =
			(struct bfa_ioctl_itnim_itnstats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_itnim_s *itnim;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->lpwwn);
	if (!fcs_port) {
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
		bfa_trc(bfad, 0);
	} else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd->rpwwn);
		if (itnim == NULL)
			iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		else {
			iocmd->status = BFA_STATUS_OK;
			bfa_fcs_itnim_stats_get(fcs_port, iocmd->rpwwn,
					&iocmd->itnstats);
		}
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

int
bfad_iocmd_fcport_enable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcport_enable(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_fcport_disable(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcport_disable(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_ioc_get_pcifn_cfg(void *cmd)
{
	struct bfa_ioctl_pcifn_cfg_s *iocmd =
				(struct bfa_ioctl_pcifn_cfg_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_query(&bfad->bfa.modules.ablk,
				&iocmd->pcifn_cfg,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_pcifn_create(void *cmd)
{
	struct bfa_ioctl_pcifn_s *iocmd = (struct bfa_ioctl_pcifn_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_pf_create(&bfad->bfa.modules.ablk,
				&iocmd->pcifn_id, iocmd->port,
				iocmd->pcifn_class, iocmd->bandwidth,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_pcifn_delete(void *cmd)
{
	struct bfa_ioctl_pcifn_s *iocmd = (struct bfa_ioctl_pcifn_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_pf_delete(&bfad->bfa.modules.ablk,
				iocmd->pcifn_id,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_pcifn_bw(void *cmd)
{
	struct bfa_ioctl_pcifn_s *iocmd = (struct bfa_ioctl_pcifn_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_pf_update(&bfad->bfa.modules.ablk,
				iocmd->pcifn_id, iocmd->bandwidth,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	bfa_trc(bfad, iocmd->status);
out:
	return 0;
}

int
bfad_iocmd_adapter_cfg_mode(void *cmd)
{
	struct bfa_ioctl_adapter_cfg_mode_s *iocmd =
			(struct bfa_ioctl_adapter_cfg_mode_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags = 0;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_adapter_config(&bfad->bfa.modules.ablk,
				iocmd->cfg.mode, iocmd->cfg.max_pf,
				iocmd->cfg.max_vf, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_port_cfg_mode(void *cmd)
{
	struct bfa_ioctl_port_cfg_mode_s *iocmd =
			(struct bfa_ioctl_port_cfg_mode_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long flags = 0;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_ablk_port_config(&bfad->bfa.modules.ablk,
				iocmd->instance, iocmd->cfg.mode,
				iocmd->cfg.max_pf, iocmd->cfg.max_vf,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_ablk_optrom(unsigned int cmd, void *pcmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)pcmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (cmd == IOCMD_FLASH_ENABLE_OPTROM)
		iocmd->status = bfa_ablk_optrom_en(&bfad->bfa.modules.ablk,
					bfad_hcb_comp, &fcomp);
	else
		iocmd->status = bfa_ablk_optrom_dis(&bfad->bfa.modules.ablk,
					bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_faa_query(void *cmd)
{
	struct bfa_ioctl_faa_attr_s *iocmd = (struct bfa_ioctl_faa_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp    fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	iocmd->status = BFA_STATUS_OK;
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_faa_query(&bfad->bfa, &iocmd->faa_attr,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_cee_attr(void *cmd)
{
	struct bfa_ioctl_cee_attr_s *iocmd =
				(struct bfa_ioctl_cee_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	struct bfad_hal_comp	cee_comp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->buf_size);
	if (!iocmd_bufptr)
		return -ENOMEM;

	cee_comp.status = 0;
	init_completion(&cee_comp.comp);
	mutex_lock(&bfad_mutex);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_cee_get_attr(&bfad->bfa.modules.cee, iocmd_bufptr,
					 bfad_hcb_comp, &cee_comp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		mutex_unlock(&bfad_mutex);
		bfa_trc(bfad, 0x5555);
		goto out;
	}
	wait_for_completion(&cee_comp.comp);
	mutex_unlock(&bfad_mutex);

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->buf_size)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_cee_get_stats(void *cmd)
{
	struct bfa_ioctl_cee_stats_s *iocmd =
				(struct bfa_ioctl_cee_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	struct bfad_hal_comp	cee_comp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->buf_size);
	if (!iocmd_bufptr)
		return -ENOMEM;

	cee_comp.status = 0;
	init_completion(&cee_comp.comp);
	mutex_lock(&bfad_mutex);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_cee_get_stats(&bfad->bfa.modules.cee, iocmd_bufptr,
					bfad_hcb_comp, &cee_comp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		mutex_unlock(&bfad_mutex);
		bfa_trc(bfad, 0x5555);
		goto out;
	}
	wait_for_completion(&cee_comp.comp);
	mutex_unlock(&bfad_mutex);

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->buf_size)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	if (iocmd_bufptr)
		vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_cee_reset_stats(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_cee_reset_stats(&bfad->bfa.modules.cee, NULL, NULL);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		bfa_trc(bfad, 0x5555);
	return 0;
}

int
bfad_iocmd_sfp_media(void *cmd)
{
	struct bfa_ioctl_sfp_media_s *iocmd =
				(struct bfa_ioctl_sfp_media_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp	fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_sfp_media(BFA_SFP_MOD(&bfad->bfa), &iocmd->media,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_SFP_NOT_READY)
		goto out;

	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_sfp_speed(void *cmd)
{
	struct bfa_ioctl_sfp_speed_s *iocmd =
				(struct bfa_ioctl_sfp_speed_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp	fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_sfp_speed(BFA_SFP_MOD(&bfad->bfa), iocmd->speed,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_SFP_NOT_READY)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_flash_get_attr(void *cmd)
{
	struct bfa_ioctl_flash_attr_s *iocmd =
			(struct bfa_ioctl_flash_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_get_attr(BFA_FLASH(&bfad->bfa), &iocmd->attr,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_flash_erase_part(void *cmd)
{
	struct bfa_ioctl_flash_s *iocmd = (struct bfa_ioctl_flash_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_erase_part(BFA_FLASH(&bfad->bfa), iocmd->type,
				iocmd->instance, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_flash_update_part(void *cmd)
{
	struct bfa_ioctl_flash_s *iocmd = (struct bfa_ioctl_flash_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->bufsz);
	if (iocmd_bufptr == NULL)
		return -ENOMEM;

	if (copy_from_user((void *)iocmd_bufptr,
		(void __user *)(size_t)iocmd->buf_ptr, iocmd->bufsz)) {
		vfree(iocmd_bufptr);
		return -EIO;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_update_part(BFA_FLASH(&bfad->bfa),
				iocmd->type, iocmd->instance, iocmd_bufptr,
				iocmd->bufsz, 0, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_flash_read_part(void *cmd)
{
	struct bfa_ioctl_flash_s *iocmd = (struct bfa_ioctl_flash_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	void	*iocmd_bufptr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->bufsz);
	if (!iocmd_bufptr) {
		bfa_trc(bfad, 0x4444);
		return -ENOMEM;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_read_part(BFA_FLASH(&bfad->bfa), iocmd->type,
				iocmd->instance, iocmd_bufptr, iocmd->bufsz, 0,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			iocmd_bufptr, iocmd->bufsz)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_diag_temp(void *cmd)
{
	struct bfa_ioctl_diag_get_temp_s *iocmd =
			(struct bfa_ioctl_diag_get_temp_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_diag_tsensor_query(BFA_DIAG_MOD(&bfad->bfa),
				&iocmd->result, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_diag_memtest(void *cmd)
{
	struct bfa_ioctl_diag_memtest_s *iocmd =
			(struct bfa_ioctl_diag_memtest_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_diag_memtest(BFA_DIAG_MOD(&bfad->bfa),
				&iocmd->memtest, iocmd->pat,
				&iocmd->result, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_diag_loopback(void *cmd)
{
	struct bfa_ioctl_diag_loopback_s *iocmd =
			(struct bfa_ioctl_diag_loopback_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcdiag_loopback(&bfad->bfa, iocmd->opmode,
				iocmd->speed, iocmd->lpcnt, iocmd->pat,
				&iocmd->result, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_diag_fwping(void *cmd)
{
	struct bfa_ioctl_diag_fwping_s *iocmd =
			(struct bfa_ioctl_diag_fwping_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_diag_fwping(BFA_DIAG_MOD(&bfad->bfa), iocmd->cnt,
				iocmd->pattern, &iocmd->result,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	bfa_trc(bfad, 0x77771);
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_diag_queuetest(void *cmd)
{
	struct bfa_ioctl_diag_qtest_s *iocmd =
				(struct bfa_ioctl_diag_qtest_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcdiag_queuetest(&bfad->bfa, iocmd->force,
				iocmd->queue, &iocmd->result,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_diag_sfp(void *cmd)
{
	struct bfa_ioctl_sfp_show_s *iocmd =
			(struct bfa_ioctl_sfp_show_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_sfp_show(BFA_SFP_MOD(&bfad->bfa), &iocmd->sfp,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	bfa_trc(bfad, iocmd->status);
out:
	return 0;
}

int
bfad_iocmd_diag_led(void *cmd)
{
	struct bfa_ioctl_diag_led_s *iocmd =
				(struct bfa_ioctl_diag_led_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_diag_ledtest(BFA_DIAG_MOD(&bfad->bfa),
				&iocmd->ledtest);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

int
bfad_iocmd_diag_beacon_lport(void *cmd)
{
	struct bfa_ioctl_diag_beacon_s *iocmd =
			(struct bfa_ioctl_diag_beacon_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_diag_beacon_port(BFA_DIAG_MOD(&bfad->bfa),
				iocmd->beacon, iocmd->link_e2e_beacon,
				iocmd->second);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

int
bfad_iocmd_diag_lb_stat(void *cmd)
{
	struct bfa_ioctl_diag_lb_stat_s *iocmd =
			(struct bfa_ioctl_diag_lb_stat_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcdiag_lb_is_running(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	bfa_trc(bfad, iocmd->status);

	return 0;
}

int
bfad_iocmd_phy_get_attr(void *cmd)
{
	struct bfa_ioctl_phy_attr_s *iocmd =
			(struct bfa_ioctl_phy_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_phy_get_attr(BFA_PHY(&bfad->bfa), iocmd->instance,
				&iocmd->attr, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_phy_get_stats(void *cmd)
{
	struct bfa_ioctl_phy_stats_s *iocmd =
			(struct bfa_ioctl_phy_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_phy_get_stats(BFA_PHY(&bfad->bfa), iocmd->instance,
				&iocmd->stats, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_phy_read(void *cmd)
{
	struct bfa_ioctl_phy_s *iocmd = (struct bfa_ioctl_phy_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	void	*iocmd_bufptr;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->bufsz);
	if (!iocmd_bufptr) {
		bfa_trc(bfad, 0x4444);
		return -ENOMEM;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_phy_read(BFA_PHY(&bfad->bfa),
				iocmd->instance, iocmd_bufptr, iocmd->bufsz,
				0, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
	if (iocmd->status != BFA_STATUS_OK)
		goto out;

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->bufsz)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_vhba_query(void *cmd)
{
	struct bfa_ioctl_vhba_attr_s *iocmd =
			(struct bfa_ioctl_vhba_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_vhba_attr_s *attr = &iocmd->attr;
	unsigned long flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	attr->pwwn =  bfad->bfa.ioc.attr->pwwn;
	attr->nwwn =  bfad->bfa.ioc.attr->nwwn;
	attr->plog_enabled = (bfa_boolean_t)bfad->bfa.plog->plog_enabled;
	attr->io_profile = bfa_fcpim_get_io_profile(&bfad->bfa);
	attr->path_tov  = bfa_fcpim_path_tov_get(&bfad->bfa);
	iocmd->status = BFA_STATUS_OK;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

int
bfad_iocmd_phy_update(void *cmd)
{
	struct bfa_ioctl_phy_s *iocmd = (struct bfa_ioctl_phy_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	iocmd_bufptr = vzalloc(iocmd->bufsz);
	if (iocmd_bufptr == NULL)
		return -ENOMEM;

	if (copy_from_user((void *)iocmd_bufptr,
		(void __user *)(size_t)iocmd->buf_ptr, iocmd->bufsz)) {
		vfree(iocmd_bufptr);
		return -EIO;
	}

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_phy_update(BFA_PHY(&bfad->bfa),
				iocmd->instance, iocmd_bufptr, iocmd->bufsz,
				0, bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_porglog_get(void *cmd)
{
	struct bfa_ioctl_debug_s *iocmd = (struct bfa_ioctl_debug_s *)cmd;
	struct bfad_s *bfad = NULL;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (iocmd->bufsz < sizeof(struct bfa_plog_s)) {
		bfa_trc(bfad, sizeof(struct bfa_plog_s));
		iocmd->status = BFA_STATUS_EINVAL;
		goto out;
	}

	iocmd->status = BFA_STATUS_OK;

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 &bfad->plog_buf, sizeof(struct bfa_plog_s)))
		return -EIO;
out:
	return 0;
}

#define BFA_DEBUG_FW_CORE_CHUNK_SZ	0x4000U /* 16K chunks for FW dump */
#define bfa_is_aligned(__val, __bytes)	(((__val) & ((8 * (__bytes)) - 1)) == 0)
int
bfad_iocmd_debug_fw_core(void *cmd)
{
	struct bfa_ioctl_debug_s *iocmd = (struct bfa_ioctl_debug_s *)cmd;
	struct bfad_s *bfad = NULL;
	void	*iocmd_bufptr;
	unsigned long	flags;
	u32 offset;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (iocmd->bufsz < BFA_DEBUG_FW_CORE_CHUNK_SZ ||
			!bfa_is_aligned(iocmd->bufsz, sizeof(u16)) ||
			!bfa_is_aligned(iocmd->offset, sizeof(u32))) {
		bfa_trc(bfad, BFA_DEBUG_FW_CORE_CHUNK_SZ);
		iocmd->status = BFA_STATUS_EINVAL;
		return 0;
	}

	iocmd_bufptr = vzalloc(iocmd->bufsz);
	if (!iocmd_bufptr) {
		bfa_trc(bfad, 0x4444);
		return -ENOMEM;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	offset = iocmd->offset;
	iocmd->status = bfa_ioc_debug_fwcore(&bfad->bfa.ioc, iocmd_bufptr,
				&offset, &iocmd->bufsz);
	iocmd->offset = offset;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, 0x5555);
		goto out;
	}

	if (copy_to_user((void __user *)(size_t)iocmd->buf_ptr,
			 iocmd_bufptr, iocmd->bufsz)) {
		vfree(iocmd_bufptr);
		bfa_trc(bfad, 0x55551);
		return -EIO;
	}
out:
	vfree(iocmd_bufptr);
	return 0;
}

int
bfad_iocmd_debug_ctl(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (v_cmd == IOCMD_DEBUG_FW_STATE_CLR) {
		spin_lock_irqsave(&bfad->bfad_lock, flags);
		bfad->bfa.ioc.dbg_fwsave_once = BFA_TRUE;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	} else if (v_cmd == IOCMD_DEBUG_PORTLOG_CLR)
		bfad->plog_buf.head = bfad->plog_buf.tail = 0;
	else if (v_cmd == IOCMD_DEBUG_START_DTRC)
		bfa_trc_init(bfad->trcmod);
	else if (v_cmd == IOCMD_DEBUG_STOP_DTRC)
		bfa_trc_stop(bfad->trcmod);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_porglog_ctl(void *cmd)
{
	struct bfa_ioctl_portlogctl_s *iocmd =
				(struct bfa_ioctl_portlogctl_s *)cmd;
	struct bfad_s *bfad = NULL;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	if (iocmd->ctl == BFA_TRUE)
		bfad->plog_buf.plog_enabled = 1;
	else
		bfad->plog_buf.plog_enabled = 0;

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_fcpim_cfg_profile(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_fcpim_profile_s *iocmd =
				(struct bfa_ioctl_fcpim_profile_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct timeval  tv;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	do_gettimeofday(&tv);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (v_cmd == IOCMD_FCPIM_PROFILE_ON)
		iocmd->status = bfa_fcpim_profile_on(&bfad->bfa, tv.tv_sec);
	else if (v_cmd == IOCMD_FCPIM_PROFILE_OFF)
		iocmd->status = bfa_fcpim_profile_off(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

static int
bfad_iocmd_itnim_get_ioprofile(void *cmd)
{
	struct bfa_ioctl_itnim_ioprofile_s *iocmd =
				(struct bfa_ioctl_itnim_ioprofile_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_itnim_s *itnim;
	unsigned long   flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs,
				iocmd->vf_id, iocmd->lpwwn);
	if (!fcs_port)
		iocmd->status = BFA_STATUS_UNKNOWN_LWWN;
	else {
		itnim = bfa_fcs_itnim_lookup(fcs_port, iocmd->rpwwn);
		if (itnim == NULL)
			iocmd->status = BFA_STATUS_UNKNOWN_RWWN;
		else
			iocmd->status = bfa_itnim_get_ioprofile(
						bfa_fcs_itnim_get_halitn(itnim),
						&iocmd->ioprofile);
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	return 0;
}

int
bfad_iocmd_fcport_get_stats(void *cmd)
{
	struct bfa_ioctl_fcport_stats_s *iocmd =
				(struct bfa_ioctl_fcport_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;
	struct bfa_cb_pending_q_s cb_qe;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	bfa_pending_q_init(&cb_qe, (bfa_cb_cbfn_t)bfad_hcb_comp,
			   &fcomp, &iocmd->stats);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcport_get_stats(&bfad->bfa, &cb_qe);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_fcport_reset_stats(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;
	struct bfa_cb_pending_q_s cb_qe;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	bfa_pending_q_init(&cb_qe, (bfa_cb_cbfn_t)bfad_hcb_comp, &fcomp, NULL);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_fcport_clear_stats(&bfad->bfa, &cb_qe);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_boot_cfg(void *cmd)
{
	struct bfa_ioctl_boot_s *iocmd = (struct bfa_ioctl_boot_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_update_part(BFA_FLASH(&bfad->bfa),
			BFA_FLASH_PART_BOOT, PCI_FUNC(bfad->pcidev->devfn),
			&iocmd->cfg, sizeof(struct bfa_boot_cfg_s), 0,
			bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_boot_query(void *cmd)
{
	struct bfa_ioctl_boot_s *iocmd = (struct bfa_ioctl_boot_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_read_part(BFA_FLASH(&bfad->bfa),
			BFA_FLASH_PART_BOOT, PCI_FUNC(bfad->pcidev->devfn),
			&iocmd->cfg, sizeof(struct bfa_boot_cfg_s), 0,
			bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_preboot_query(void *cmd)
{
	struct bfa_ioctl_preboot_s *iocmd = (struct bfa_ioctl_preboot_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfi_iocfc_cfgrsp_s *cfgrsp;
	struct bfa_boot_pbc_s *pbcfg = &iocmd->cfg;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	cfgrsp = bfad->bfa.iocfc.cfgrsp;
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	pbcfg->enable = cfgrsp->pbc_cfg.boot_enabled;
	pbcfg->nbluns = cfgrsp->pbc_cfg.nbluns;
	pbcfg->speed = cfgrsp->pbc_cfg.port_speed;
	memcpy(pbcfg->pblun, cfgrsp->pbc_cfg.blun, sizeof(pbcfg->pblun));
	iocmd->status = BFA_STATUS_OK;
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return 0;
}

int
bfad_iocmd_ethboot_cfg(void *cmd)
{
	struct bfa_ioctl_ethboot_s *iocmd = (struct bfa_ioctl_ethboot_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_update_part(BFA_FLASH(&bfad->bfa),
				BFA_FLASH_PART_PXECFG,
				bfad->bfa.ioc.port_id, &iocmd->cfg,
				sizeof(struct bfa_ethboot_cfg_s), 0,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_ethboot_query(void *cmd)
{
	struct bfa_ioctl_ethboot_s *iocmd = (struct bfa_ioctl_ethboot_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->status = bfa_flash_read_part(BFA_FLASH(&bfad->bfa),
				BFA_FLASH_PART_PXECFG,
				bfad->bfa.ioc.port_id, &iocmd->cfg,
				sizeof(struct bfa_ethboot_cfg_s), 0,
				bfad_hcb_comp, &fcomp);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK)
		goto out;
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_cfg_trunk(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	struct bfa_fcport_trunk_s *trunk;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	trunk = &fcport->trunk;
	spin_lock_irqsave(&bfad->bfad_lock, flags);

	if (v_cmd == IOCMD_TRUNK_ENABLE) {
		trunk->attr.state = BFA_TRUNK_OFFLINE;
		bfa_fcport_disable(&bfad->bfa);
		fcport->cfg.trunked = BFA_TRUE;
	} else if (v_cmd == IOCMD_TRUNK_DISABLE) {
		trunk->attr.state = BFA_TRUNK_DISABLED;
		bfa_fcport_disable(&bfad->bfa);
		fcport->cfg.trunked = BFA_FALSE;
	}

	if (!bfa_fcport_is_disabled(&bfad->bfa))
		bfa_fcport_enable(&bfad->bfa);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_trunk_get_attr(void *cmd)
{
	struct bfa_ioctl_trunk_attr_s *iocmd =
				(struct bfa_ioctl_trunk_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	struct bfa_fcport_trunk_s *trunk;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	trunk = &fcport->trunk;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	memcpy((void *)&iocmd->attr, (void *)&trunk->attr,
		sizeof(struct bfa_trunk_attr_s));
	iocmd->attr.port_id = bfa_lps_get_base_pid(&bfad->bfa);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_qos(void *cmd, unsigned int v_cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	if (bfa_ioc_get_type(&bfad->bfa.ioc) == BFA_IOC_TYPE_FC) {
		if (v_cmd == IOCMD_QOS_ENABLE)
			fcport->cfg.qos_enabled = BFA_TRUE;
		else if (v_cmd == IOCMD_QOS_DISABLE)
			fcport->cfg.qos_enabled = BFA_FALSE;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_qos_get_attr(void *cmd)
{
	struct bfa_ioctl_qos_attr_s *iocmd = (struct bfa_ioctl_qos_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->attr.state = fcport->qos_attr.state;
	iocmd->attr.total_bb_cr = be32_to_cpu(fcport->qos_attr.total_bb_cr);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_qos_get_vc_attr(void *cmd)
{
	struct bfa_ioctl_qos_vc_attr_s *iocmd =
				(struct bfa_ioctl_qos_vc_attr_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcport_s *fcport;
	struct bfa_qos_vc_attr_s *bfa_vc_attr;
	unsigned long	flags;
	u32	i = 0;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	fcport = BFA_FCPORT_MOD(&bfad->bfa);
	bfa_vc_attr = &fcport->qos_vc_attr;
	spin_lock_irqsave(&bfad->bfad_lock, flags);
	iocmd->attr.total_vc_count = be16_to_cpu(bfa_vc_attr->total_vc_count);
	iocmd->attr.shared_credit  = be16_to_cpu(bfa_vc_attr->shared_credit);
	iocmd->attr.elp_opmode_flags  =
				be32_to_cpu(bfa_vc_attr->elp_opmode_flags);

	/* Individual VC info */
	while (i < iocmd->attr.total_vc_count) {
		iocmd->attr.vc_info[i].vc_credit =
				bfa_vc_attr->vc_info[i].vc_credit;
		iocmd->attr.vc_info[i].borrow_credit =
				bfa_vc_attr->vc_info[i].borrow_credit;
		iocmd->attr.vc_info[i].priority =
				bfa_vc_attr->vc_info[i].priority;
		i++;
	}
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	iocmd->status = BFA_STATUS_OK;
	return 0;
}

int
bfad_iocmd_qos_get_stats(void *cmd)
{
	struct bfa_ioctl_fcport_stats_s *iocmd =
				(struct bfa_ioctl_fcport_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;
	struct bfa_cb_pending_q_s cb_qe;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	bfa_pending_q_init(&cb_qe, (bfa_cb_cbfn_t)bfad_hcb_comp,
			   &fcomp, &iocmd->stats);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	WARN_ON(!bfa_ioc_get_fcmode(&bfad->bfa.ioc));
	iocmd->status = bfa_fcport_get_stats(&bfad->bfa, &cb_qe);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_qos_reset_stats(void *cmd)
{
	struct bfa_ioctl_gen_s *iocmd = (struct bfa_ioctl_gen_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfad_hal_comp fcomp;
	unsigned long	flags;
	struct bfa_cb_pending_q_s cb_qe;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	init_completion(&fcomp.comp);
	bfa_pending_q_init(&cb_qe, (bfa_cb_cbfn_t)bfad_hcb_comp,
			   &fcomp, NULL);

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	WARN_ON(!bfa_ioc_get_fcmode(&bfad->bfa.ioc));
	iocmd->status = bfa_fcport_clear_stats(&bfad->bfa, &cb_qe);
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	if (iocmd->status != BFA_STATUS_OK) {
		bfa_trc(bfad, iocmd->status);
		goto out;
	}
	wait_for_completion(&fcomp.comp);
	iocmd->status = fcomp.status;
out:
	return 0;
}

int
bfad_iocmd_vf_get_stats(void *cmd)
{
	struct bfa_ioctl_vf_stats_s *iocmd =
			(struct bfa_ioctl_vf_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_fabric_s	*fcs_vf;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd->vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}
	memcpy((void *)&iocmd->stats, (void *)&fcs_vf->stats,
		sizeof(struct bfa_vf_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

int
bfad_iocmd_vf_clr_stats(void *cmd)
{
	struct bfa_ioctl_vf_reset_stats_s *iocmd =
			(struct bfa_ioctl_vf_reset_stats_s *)cmd;
	struct bfad_s *bfad = NULL;
	struct bfa_fcs_fabric_s	*fcs_vf;
	unsigned long	flags;

	bfad = bfad_find_bfad_by_inst_no(iocmd->bfad_num);
	if (bfad == NULL)
		return -EINVAL;

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_vf = bfa_fcs_vf_lookup(&bfad->bfa_fcs, iocmd->vf_id);
	if (fcs_vf == NULL) {
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		iocmd->status = BFA_STATUS_UNKNOWN_VFID;
		goto out;
	}
	memset((void *)&fcs_vf->stats, 0, sizeof(struct bfa_vf_stats_s));
	spin_unlock_irqrestore(&bfad->bfad_lock, flags);
	iocmd->status = BFA_STATUS_OK;
out:
	return 0;
}

/* FC passthru call backs */
u64
bfad_fcxp_get_req_sgaddr_cb(void *bfad_fcxp, int sgeid)
{
	struct bfad_fcxp	*drv_fcxp = bfad_fcxp;
	struct bfa_sge_s	*sge;
	u64	addr;

	sge = drv_fcxp->req_sge + sgeid;
	addr = (u64)(size_t) sge->sg_addr;
	return addr;
}

u32
bfad_fcxp_get_req_sglen_cb(void *bfad_fcxp, int sgeid)
{
	struct bfad_fcxp	*drv_fcxp = bfad_fcxp;
	struct bfa_sge_s	*sge;

	sge = drv_fcxp->req_sge + sgeid;
	return sge->sg_len;
}

u64
bfad_fcxp_get_rsp_sgaddr_cb(void *bfad_fcxp, int sgeid)
{
	struct bfad_fcxp	*drv_fcxp = bfad_fcxp;
	struct bfa_sge_s	*sge;
	u64	addr;

	sge = drv_fcxp->rsp_sge + sgeid;
	addr = (u64)(size_t) sge->sg_addr;
	return addr;
}

u32
bfad_fcxp_get_rsp_sglen_cb(void *bfad_fcxp, int sgeid)
{
	struct bfad_fcxp	*drv_fcxp = bfad_fcxp;
	struct bfa_sge_s	*sge;

	sge = drv_fcxp->rsp_sge + sgeid;
	return sge->sg_len;
}

void
bfad_send_fcpt_cb(void *bfad_fcxp, struct bfa_fcxp_s *fcxp, void *cbarg,
		bfa_status_t req_status, u32 rsp_len, u32 resid_len,
		struct fchs_s *rsp_fchs)
{
	struct bfad_fcxp *drv_fcxp = bfad_fcxp;

	drv_fcxp->req_status = req_status;
	drv_fcxp->rsp_len = rsp_len;

	/* bfa_fcxp will be automatically freed by BFA */
	drv_fcxp->bfa_fcxp = NULL;
	complete(&drv_fcxp->comp);
}

struct bfad_buf_info *
bfad_fcxp_map_sg(struct bfad_s *bfad, void *payload_kbuf,
		 u32 payload_len, u32 *num_sgles)
{
	struct bfad_buf_info	*buf_base, *buf_info;
	struct bfa_sge_s	*sg_table;
	int sge_num = 1;

	buf_base = kzalloc((sizeof(struct bfad_buf_info) +
			    sizeof(struct bfa_sge_s)) * sge_num, GFP_KERNEL);
	if (!buf_base)
		return NULL;

	sg_table = (struct bfa_sge_s *) (((u8 *)buf_base) +
			(sizeof(struct bfad_buf_info) * sge_num));

	/* Allocate dma coherent memory */
	buf_info = buf_base;
	buf_info->size = payload_len;
	buf_info->virt = dma_alloc_coherent(&bfad->pcidev->dev, buf_info->size,
					&buf_info->phys, GFP_KERNEL);
	if (!buf_info->virt)
		goto out_free_mem;

	/* copy the linear bsg buffer to buf_info */
	memset(buf_info->virt, 0, buf_info->size);
	if (payload_kbuf) {
		if (copy_from_user((u8 *) buf_info->virt,
				(void __user *)payload_kbuf, payload_len)) {
			dma_free_coherent(&bfad->pcidev->dev, buf_info->size,
					  buf_info->virt, buf_info->phys);
			goto out_free_mem;
		}
	}

	/*
	 * Setup SG table
	 */
	sg_table->sg_len = buf_info->size;
	sg_table->sg_addr = (void *)(size_t) buf_info->phys;

	*num_sgles = sge_num;

	return buf_base;

out_free_mem:
	kfree(buf_base);
	return NULL;
}

void
bfad_fcxp_free_mem(struct bfad_s *bfad, struct bfad_buf_info *buf_base,
		   u32 num_sgles)
{
	int i;
	struct bfad_buf_info *buf_info = buf_base;

	if (buf_base) {
		for (i = 0; i < num_sgles; buf_info++, i++) {
			if (buf_info->virt != NULL)
				dma_free_coherent(&bfad->pcidev->dev,
					buf_info->size, buf_info->virt,
					buf_info->phys);
		}
		kfree(buf_base);
	}
}

int
bfad_fcxp_iocmd_send(struct bfad_fcxp *drv_fcxp, bfa_ioctl_fcpt_t *iocmd_fcpt)
{
	struct bfa_fcxp_s *hal_fcxp;
	struct bfad_s   *bfad = drv_fcxp->port->bfad;
	unsigned long   flags;
	u8 lp_tag;

	spin_lock_irqsave(&bfad->bfad_lock, flags);

	/* Allocate bfa_fcxp structure */
	hal_fcxp = bfa_fcxp_alloc(drv_fcxp, &bfad->bfa,
				  drv_fcxp->num_req_sgles,
				  drv_fcxp->num_rsp_sgles,
				  bfad_fcxp_get_req_sgaddr_cb,
				  bfad_fcxp_get_req_sglen_cb,
				  bfad_fcxp_get_rsp_sgaddr_cb,
				  bfad_fcxp_get_rsp_sglen_cb);
	if (!hal_fcxp) {
		bfa_trc(bfad, 0);
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		return BFA_STATUS_ENOMEM;
	}

	drv_fcxp->bfa_fcxp = hal_fcxp;

	lp_tag = bfa_lps_get_tag_from_pid(&bfad->bfa, iocmd_fcpt->fchs.s_id);

	bfa_fcxp_send(hal_fcxp, drv_fcxp->bfa_rport, iocmd_fcpt->vf_id, lp_tag,
		      iocmd_fcpt->cts, iocmd_fcpt->cos,
		      iocmd_fcpt->reqsz,
		      &iocmd_fcpt->fchs, bfad_send_fcpt_cb, bfad,
		      iocmd_fcpt->rspsz, iocmd_fcpt->tsecs);

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	return BFA_STATUS_OK;
}

int
bfad_iocmd_els_ct_request(unsigned long arg)
{
	struct bfad_s *bfad;
	bfa_ioctl_fcpt_t *iocmd_fcpt;
	struct bfad_fcxp    *drv_fcxp;
	struct bfa_fcs_lport_s *fcs_port;
	struct bfa_fcs_rport_s *fcs_rport;
	unsigned long flags;
	int rc = -EINVAL;

	/*
	 * Allocate buffer for iocmd_fcpt and do a copy_from_user op for payload
	 * buffer of size bfa_ioctl_fcpt_t
	 */
	iocmd_fcpt = kzalloc(sizeof(bfa_ioctl_fcpt_t), GFP_KERNEL);
	if (!iocmd_fcpt) {
		rc = -ENOMEM;
		goto out;
	}

	if (copy_from_user((u8 *)iocmd_fcpt, (void __user *)arg,
				sizeof(bfa_ioctl_fcpt_t))) {
		kfree(iocmd_fcpt);
		rc = -EIO;
		goto out;
	}

	bfad = bfad_find_bfad_by_inst_no(iocmd_fcpt->bfad_num);
	if (bfad == NULL) {
		kfree(iocmd_fcpt);
		return -EINVAL;
	}

	drv_fcxp = kzalloc(sizeof(struct bfad_fcxp), GFP_KERNEL);
	if (drv_fcxp == NULL) {
		kfree(iocmd_fcpt);
		rc = -ENOMEM;
		goto out;
	}

	spin_lock_irqsave(&bfad->bfad_lock, flags);
	fcs_port = bfa_fcs_lookup_port(&bfad->bfa_fcs, iocmd_fcpt->vf_id,
				       iocmd_fcpt->lpwwn);
	if (fcs_port == NULL) {
		iocmd_fcpt->status = BFA_STATUS_UNKNOWN_LWWN;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		goto out_free_mem;
	}

	/* Check if the port is online before sending FC Passthru cmd */
	if (!bfa_fcs_lport_is_online(fcs_port)) {
		iocmd_fcpt->status = BFA_STATUS_PORT_OFFLINE;
		spin_unlock_irqrestore(&bfad->bfad_lock, flags);
		goto out_free_mem;
	}

	drv_fcxp->port = fcs_port->bfad_port;

	if (drv_fcxp->port->bfad == 0)
		drv_fcxp->port->bfad = bfad;

	/* Fetch the bfa_rport - if nexus needed */
	if (iocmd_fcpt->nport_nexus_needed) {
		/* BSG RPT commands: nexus needed */
		fcs_rport = bfa_fcs_lport_get_rport_by_pwwn(fcs_port,
						iocmd_fcpt->dpwwn);
		if (fcs_rport == NULL) {
			iocmd_fcpt->status = BFA_STATUS_UNKNOWN_RWWN;
			spin_unlock_irqrestore(&bfad->bfad_lock, flags);
			goto out_free_mem;
		}

		drv_fcxp->bfa_rport = fcs_rport->bfa_rport;

	} else {
		/* BSG HST commands: no nexus needed */
		drv_fcxp->bfa_rport = NULL;
	}

	spin_unlock_irqrestore(&bfad->bfad_lock, flags);

	/* allocate memory for req / rsp buffers */
	drv_fcxp->reqbuf_info = bfad_fcxp_map_sg(bfad,
					(char *)(size_t)iocmd_fcpt->reqbuf_ptr,
					iocmd_fcpt->reqsz,
					&drv_fcxp->num_req_sgles);
	if (!drv_fcxp->reqbuf_info) {
		printk(KERN_INFO "bfa %s: fcpt request fcxp_map_sg failed\n",
				bfad->pci_name);
		rc = -ENOMEM;
		goto out_free_mem;
	}

	drv_fcxp->req_sge = (struct bfa_sge_s *)
			    (((u8 *)drv_fcxp->reqbuf_info) +
			    (sizeof(struct bfad_buf_info) *
					drv_fcxp->num_req_sgles));

	/* map rsp sg */
	drv_fcxp->rspbuf_info = bfad_fcxp_map_sg(bfad, NULL,
					iocmd_fcpt->rspsz,
					&drv_fcxp->num_rsp_sgles);
	if (!drv_fcxp->rspbuf_info) {
		printk(KERN_INFO "bfa %s: fcpt response fcxp_map_sg failed\n",
				bfad->pci_name);
		rc = -ENOMEM;
		goto out_free_mem;
	}

	drv_fcxp->rsp_sge = (struct bfa_sge_s  *)
			    (((u8 *)drv_fcxp->rspbuf_info) +
			    (sizeof(struct bfad_buf_info) *
					drv_fcxp->num_rsp_sgles));

	/* fcxp send */
	init_completion(&drv_fcxp->comp);
	rc = bfad_fcxp_iocmd_send(drv_fcxp, iocmd_fcpt);
	if (rc == BFA_STATUS_OK) {
		wait_for_completion(&drv_fcxp->comp);
		if (drv_fcxp->req_status == BFA_STATUS_OK) {
			if (copy_to_user((char *)(size_t)iocmd_fcpt->rspbuf_ptr,
					(u8 *)((struct bfad_buf_info *)
					 drv_fcxp->rspbuf_info)->virt,
					 iocmd_fcpt->rspsz)) {
				rc = -EIO;
				goto out_free_mem;
			}
		}
		iocmd_fcpt->status = drv_fcxp->req_status;
	} else {
		iocmd_fcpt->status = rc;
		goto out_free_mem;
	}

out_free_mem:
	bfad_fcxp_free_mem(bfad, drv_fcxp->rspbuf_info,
			   drv_fcxp->num_rsp_sgles);
	bfad_fcxp_free_mem(bfad, drv_fcxp->reqbuf_info,
			   drv_fcxp->num_req_sgles);

	/* Need a copy to user op */
	if (copy_to_user((void __user *)arg, (void *) iocmd_fcpt,
			 sizeof(bfa_ioctl_fcpt_t)))
		rc = -EIO;

	kfree(iocmd_fcpt);
	kfree(drv_fcxp);
out:
	return rc;
}

static int
bfad_iocmd_handler(unsigned int cmd, void *iocmd, unsigned long arg)
{
	int rc = -EINVAL;

	switch (cmd) {
	case IOCMD_IOC_ENABLE:
	case IOCMD_VHBA_ENABLE:
		rc = bfad_iocmd_ioc_enable(iocmd);
		break;
	case IOCMD_IOC_DISABLE:
	case IOCMD_VHBA_DISABLE:
		rc = bfad_iocmd_ioc_disable(iocmd);
		break;
	case IOCMD_IOC_GET_INFO:
		rc = bfad_iocmd_ioc_get_info(iocmd);
		break;
	case IOCMD_IOC_GET_ATTR:
		rc = bfad_iocmd_ioc_get_attr(iocmd);
		break;
	case IOCMD_IOC_GET_STATS:
		rc = bfad_iocmd_ioc_get_stats(iocmd);
		break;
	case IOCMD_IOC_GET_FWSTATS:
		rc = bfad_iocmd_ioc_get_fwstats(iocmd);
		break;
	case IOCMD_IOC_RESET_STATS:
	case IOCMD_IOC_RESET_FWSTATS:
		rc = bfad_iocmd_ioc_reset_stats(iocmd, cmd);
		break;
	case IOCMD_IOC_SET_ADAPTER_NAME:
	case IOCMD_IOC_SET_PORT_NAME:
		rc = bfad_iocmd_ioc_set_name(iocmd, cmd);
		break;
	case IOCMD_IOCFC_GET_ATTR:
		rc = bfad_iocmd_iocfc_get_attr(iocmd);
		break;
	case IOCMD_IOCFC_SET_INTR:
		rc = bfad_iocmd_iocfc_set_intr(iocmd);
		break;
	case IOCMD_PORT_ENABLE:
		rc = bfad_iocmd_port_enable(iocmd);
		break;
	case IOCMD_PORT_DISABLE:
		rc = bfad_iocmd_port_disable(iocmd);
		break;
	case IOCMD_PORT_GET_ATTR:
		rc = bfad_iocmd_port_get_attr(iocmd);
		break;
	case IOCMD_PORT_GET_STATS:
		rc = bfad_iocmd_port_get_stats(iocmd);
		break;
	case IOCMD_PORT_RESET_STATS:
		rc = bfad_iocmd_port_reset_stats(iocmd);
		break;
	case IOCMD_PORT_CFG_TOPO:
	case IOCMD_PORT_CFG_SPEED:
	case IOCMD_PORT_CFG_ALPA:
	case IOCMD_PORT_CLR_ALPA:
		rc = bfad_iocmd_set_port_cfg(iocmd, cmd);
		break;
	case IOCMD_PORT_CFG_MAXFRSZ:
		rc = bfad_iocmd_port_cfg_maxfrsize(iocmd);
		break;
	case IOCMD_PORT_BBSC_ENABLE:
	case IOCMD_PORT_BBSC_DISABLE:
		rc = bfad_iocmd_port_cfg_bbsc(iocmd, cmd);
		break;
	case IOCMD_LPORT_GET_ATTR:
		rc = bfad_iocmd_lport_get_attr(iocmd);
		break;
	case IOCMD_LPORT_GET_STATS:
		rc = bfad_iocmd_lport_get_stats(iocmd);
		break;
	case IOCMD_LPORT_RESET_STATS:
		rc = bfad_iocmd_lport_reset_stats(iocmd);
		break;
	case IOCMD_LPORT_GET_IOSTATS:
		rc = bfad_iocmd_lport_get_iostats(iocmd);
		break;
	case IOCMD_LPORT_GET_RPORTS:
		rc = bfad_iocmd_lport_get_rports(iocmd);
		break;
	case IOCMD_RPORT_GET_ATTR:
		rc = bfad_iocmd_rport_get_attr(iocmd);
		break;
	case IOCMD_RPORT_GET_ADDR:
		rc = bfad_iocmd_rport_get_addr(iocmd);
		break;
	case IOCMD_RPORT_GET_STATS:
		rc = bfad_iocmd_rport_get_stats(iocmd);
		break;
	case IOCMD_RPORT_RESET_STATS:
		rc = bfad_iocmd_rport_clr_stats(iocmd);
		break;
	case IOCMD_RPORT_SET_SPEED:
		rc = bfad_iocmd_rport_set_speed(iocmd);
		break;
	case IOCMD_VPORT_GET_ATTR:
		rc = bfad_iocmd_vport_get_attr(iocmd);
		break;
	case IOCMD_VPORT_GET_STATS:
		rc = bfad_iocmd_vport_get_stats(iocmd);
		break;
	case IOCMD_VPORT_RESET_STATS:
		rc = bfad_iocmd_vport_clr_stats(iocmd);
		break;
	case IOCMD_FABRIC_GET_LPORTS:
		rc = bfad_iocmd_fabric_get_lports(iocmd);
		break;
	case IOCMD_RATELIM_ENABLE:
	case IOCMD_RATELIM_DISABLE:
		rc = bfad_iocmd_ratelim(cmd, iocmd);
		break;
	case IOCMD_RATELIM_DEF_SPEED:
		rc = bfad_iocmd_ratelim_speed(cmd, iocmd);
		break;
	case IOCMD_FCPIM_FAILOVER:
		rc = bfad_iocmd_cfg_fcpim(iocmd);
		break;
	case IOCMD_FCPIM_MODSTATS:
		rc = bfad_iocmd_fcpim_get_modstats(iocmd);
		break;
	case IOCMD_FCPIM_MODSTATSCLR:
		rc = bfad_iocmd_fcpim_clr_modstats(iocmd);
		break;
	case IOCMD_FCPIM_DEL_ITN_STATS:
		rc = bfad_iocmd_fcpim_get_del_itn_stats(iocmd);
		break;
	case IOCMD_ITNIM_GET_ATTR:
		rc = bfad_iocmd_itnim_get_attr(iocmd);
		break;
	case IOCMD_ITNIM_GET_IOSTATS:
		rc = bfad_iocmd_itnim_get_iostats(iocmd);
		break;
	case IOCMD_ITNIM_RESET_STATS:
		rc = bfad_iocmd_itnim_reset_stats(iocmd);
		break;
	case IOCMD_ITNIM_GET_ITNSTATS:
		rc = bfad_iocmd_itnim_get_itnstats(iocmd);
		break;
	case IOCMD_FCPORT_ENABLE:
		rc = bfad_iocmd_fcport_enable(iocmd);
		break;
	case IOCMD_FCPORT_DISABLE:
		rc = bfad_iocmd_fcport_disable(iocmd);
		break;
	case IOCMD_IOC_PCIFN_CFG:
		rc = bfad_iocmd_ioc_get_pcifn_cfg(iocmd);
		break;
	case IOCMD_PCIFN_CREATE:
		rc = bfad_iocmd_pcifn_create(iocmd);
		break;
	case IOCMD_PCIFN_DELETE:
		rc = bfad_iocmd_pcifn_delete(iocmd);
		break;
	case IOCMD_PCIFN_BW:
		rc = bfad_iocmd_pcifn_bw(iocmd);
		break;
	case IOCMD_ADAPTER_CFG_MODE:
		rc = bfad_iocmd_adapter_cfg_mode(iocmd);
		break;
	case IOCMD_PORT_CFG_MODE:
		rc = bfad_iocmd_port_cfg_mode(iocmd);
		break;
	case IOCMD_FLASH_ENABLE_OPTROM:
	case IOCMD_FLASH_DISABLE_OPTROM:
		rc = bfad_iocmd_ablk_optrom(cmd, iocmd);
		break;
	case IOCMD_FAA_QUERY:
		rc = bfad_iocmd_faa_query(iocmd);
		break;
	case IOCMD_CEE_GET_ATTR:
		rc = bfad_iocmd_cee_attr(iocmd);
		break;
	case IOCMD_CEE_GET_STATS:
		rc = bfad_iocmd_cee_get_stats(iocmd);
		break;
	case IOCMD_CEE_RESET_STATS:
		rc = bfad_iocmd_cee_reset_stats(iocmd);
		break;
	case IOCMD_SFP_MEDIA:
		rc = bfad_iocmd_sfp_media(iocmd);
		 break;
	case IOCMD_SFP_SPEED:
		rc = bfad_iocmd_sfp_speed(iocmd);
		break;
	case IOCMD_FLASH_GET_ATTR:
		rc = bfad_iocmd_flash_get_attr(iocmd);
		break;
	case IOCMD_FLASH_ERASE_PART:
		rc = bfad_iocmd_flash_erase_part(iocmd);
		break;
	case IOCMD_FLASH_UPDATE_PART:
		rc = bfad_iocmd_flash_update_part(iocmd);
		break;
	case IOCMD_FLASH_READ_PART:
		rc = bfad_iocmd_flash_read_part(iocmd);
		break;
	case IOCMD_DIAG_TEMP:
		rc = bfad_iocmd_diag_temp(iocmd);
		break;
	case IOCMD_DIAG_MEMTEST:
		rc = bfad_iocmd_diag_memtest(iocmd);
		break;
	case IOCMD_DIAG_LOOPBACK:
		rc = bfad_iocmd_diag_loopback(iocmd);
		break;
	case IOCMD_DIAG_FWPING:
		rc = bfad_iocmd_diag_fwping(iocmd);
		break;
	case IOCMD_DIAG_QUEUETEST:
		rc = bfad_iocmd_diag_queuetest(iocmd);
		break;
	case IOCMD_DIAG_SFP:
		rc = bfad_iocmd_diag_sfp(iocmd);
		break;
	case IOCMD_DIAG_LED:
		rc = bfad_iocmd_diag_led(iocmd);
		break;
	case IOCMD_DIAG_BEACON_LPORT:
		rc = bfad_iocmd_diag_beacon_lport(iocmd);
		break;
	case IOCMD_DIAG_LB_STAT:
		rc = bfad_iocmd_diag_lb_stat(iocmd);
		break;
	case IOCMD_PHY_GET_ATTR:
		rc = bfad_iocmd_phy_get_attr(iocmd);
		break;
	case IOCMD_PHY_GET_STATS:
		rc = bfad_iocmd_phy_get_stats(iocmd);
		break;
	case IOCMD_PHY_UPDATE_FW:
		rc = bfad_iocmd_phy_update(iocmd);
		break;
	case IOCMD_PHY_READ_FW:
		rc = bfad_iocmd_phy_read(iocmd);
		break;
	case IOCMD_VHBA_QUERY:
		rc = bfad_iocmd_vhba_query(iocmd);
		break;
	case IOCMD_DEBUG_PORTLOG:
		rc = bfad_iocmd_porglog_get(iocmd);
		break;
	case IOCMD_DEBUG_FW_CORE:
		rc = bfad_iocmd_debug_fw_core(iocmd);
		break;
	case IOCMD_DEBUG_FW_STATE_CLR:
	case IOCMD_DEBUG_PORTLOG_CLR:
	case IOCMD_DEBUG_START_DTRC:
	case IOCMD_DEBUG_STOP_DTRC:
		rc = bfad_iocmd_debug_ctl(iocmd, cmd);
		break;
	case IOCMD_DEBUG_PORTLOG_CTL:
		rc = bfad_iocmd_porglog_ctl(iocmd);
		break;
	case IOCMD_FCPIM_PROFILE_ON:
	case IOCMD_FCPIM_PROFILE_OFF:
		rc = bfad_iocmd_fcpim_cfg_profile(iocmd, cmd);
		break;
	case IOCMD_ITNIM_GET_IOPROFILE:
		rc = bfad_iocmd_itnim_get_ioprofile(iocmd);
		break;
	case IOCMD_FCPORT_GET_STATS:
		rc = bfad_iocmd_fcport_get_stats(iocmd);
		break;
	case IOCMD_FCPORT_RESET_STATS:
		rc = bfad_iocmd_fcport_reset_stats(iocmd);
		break;
	case IOCMD_BOOT_CFG:
		rc = bfad_iocmd_boot_cfg(iocmd);
		break;
	case IOCMD_BOOT_QUERY:
		rc = bfad_iocmd_boot_query(iocmd);
		break;
	case IOCMD_PREBOOT_QUERY:
		rc = bfad_iocmd_preboot_query(iocmd);
		break;
	case IOCMD_ETHBOOT_CFG:
		rc = bfad_iocmd_ethboot_cfg(iocmd);
		break;
	case IOCMD_ETHBOOT_QUERY:
		rc = bfad_iocmd_ethboot_query(iocmd);
		break;
	case IOCMD_TRUNK_ENABLE:
	case IOCMD_TRUNK_DISABLE:
		rc = bfad_iocmd_cfg_trunk(iocmd, cmd);
		break;
	case IOCMD_TRUNK_GET_ATTR:
		rc = bfad_iocmd_trunk_get_attr(iocmd);
		break;
	case IOCMD_QOS_ENABLE:
	case IOCMD_QOS_DISABLE:
		rc = bfad_iocmd_qos(iocmd, cmd);
		break;
	case IOCMD_QOS_GET_ATTR:
		rc = bfad_iocmd_qos_get_attr(iocmd);
		break;
	case IOCMD_QOS_GET_VC_ATTR:
		rc = bfad_iocmd_qos_get_vc_attr(iocmd);
		break;
	case IOCMD_QOS_GET_STATS:
		rc = bfad_iocmd_qos_get_stats(iocmd);
		break;
	case IOCMD_QOS_RESET_STATS:
		rc = bfad_iocmd_qos_reset_stats(iocmd);
		break;
	case IOCMD_VF_GET_STATS:
		rc = bfad_iocmd_vf_get_stats(iocmd);
		break;
	case IOCMD_VF_RESET_STATS:
		rc = bfad_iocmd_vf_clr_stats(iocmd);
		break;
	default:
		rc = -EINVAL;
		break;
	}
	return rc;
}

int
bfad_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	   unsigned long arg)
{
	int rc = -EINVAL, size = _IOC_SIZE(cmd);
	void *iocmd = NULL;

	if (!arg)
		return -EINVAL;

	/* Check if the pci probe scan is done */
	if (!bfad_is_ready())
		return -EAGAIN;

	/* Handle ELS/CT passthru requests */
	if (cmd == IOCMD_SEND_FC) {
		rc = bfad_iocmd_els_ct_request(arg);
		goto ext;
	}

	/* Allocate memory for the userspace iocmd */
	iocmd = kzalloc(size, GFP_KERNEL);
	if (!iocmd)
		return -EIO;

	/* Copy the received buffer from the user space into kernel buffer */
	if (copy_from_user((void *)iocmd, (void __user *)arg, size)) {
		kfree(iocmd);
		return -EIO;
	}

	/* Invoke common IOCMD request handler */
	rc = bfad_iocmd_handler(cmd, iocmd, arg);

	/* Copy the kernel buffer back to user space */
	if (copy_to_user((void __user *)arg, (void *)iocmd, size)) {
		kfree(iocmd);
		return -EIO;
	}

	kfree(iocmd);
ext:
	return rc;
}

static int
bfad_open(struct inode *inode, struct file *file)
{
	struct bfad_s	*bfad;
	u32	bfad_minor;

	/* Check if the pci probe scan is done */
	if (!bfad_is_ready())
		return -EAGAIN;

	bfad_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING "%s:Invalid minor device=%d for file=%p\n",
				    __func__, bfad_minor, file);
		return -EINVAL;
	}

	mutex_lock(&bfad_mutex);
	bfad->ref_count++;
	mutex_unlock(&bfad_mutex);
	return 0;
}

static int
bfad_close(struct inode *inode, struct file *file)
{
	struct bfad_s	*bfad;
	u32		bfad_minor;

	/* Check if the pci probe scan is done */
	if (!bfad_is_ready())
		return -EPERM;

	bfad_minor = MINOR(inode->i_rdev);
	bfad = bfad_find_bfad_by_inst_no(bfad_minor);
	if (bfad == NULL) {
		printk(KERN_WARNING "%s:Invalid minor device=%d for file=%p\n",
				    __func__, bfad_minor, file);
		return -EINVAL;
	}

	mutex_lock(&bfad_mutex);
	if (--bfad->ref_count) {
		mutex_unlock(&bfad_mutex);
		return 0;
	}
	mutex_unlock(&bfad_mutex);
	return 0;
}

static long
bfad_compat_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	return bfad_ioctl(file->f_dentry->d_inode, file, cmd, arg);
}

/*
 * BFA driver file operations.
 */
static const struct file_operations bfad_fops = {
	.owner = THIS_MODULE,
	.ioctl = bfad_ioctl,
	.compat_ioctl = bfad_compat_ioctl,
	.open = bfad_open,
	.release = bfad_close,
};

static int bfad_major;
static struct class *bfad_class;
static char *bfad_dev_name = BFAD_DRIVER_NAME;

int
bfad_ioctl_init(void)
{
	struct class_device *cd;

	bfad_class = class_create(THIS_MODULE, bfad_dev_name);
	if (IS_ERR(bfad_class)) {
		printk(KERN_WARNING "Unable to create sysfs class for bfa.\n");
		bfad_class = NULL;
		return BFA_STATUS_FAILED;
	}

	bfad_major = register_chrdev(0, bfad_dev_name, &bfad_fops);
	if (bfad_major < 0) {
		printk(KERN_WARNING "Unable to register chrdev for bfa.\n");
		class_destroy(bfad_class);
		bfad_class = NULL;
		return BFA_STATUS_FAILED;
	}

	cd = class_device_create(bfad_class, NULL,
			MKDEV(bfad_major, 0), NULL, "bfa");
	if (IS_ERR(cd))
		return BFA_STATUS_FAILED;

	return BFA_STATUS_OK;
}

int
bfad_ioctl_exit(void)
{
	if (!bfad_class)
		return BFA_STATUS_FAILED;

	class_device_destroy(bfad_class, MKDEV(bfad_major, 0));
	unregister_chrdev(bfad_major, bfad_dev_name);
	class_destroy(bfad_class);
	return 0;
}
