
#include <xen/types.h>
#include <xen/kernel.h>
#include <asm/cpufeature.h>

#define bit(x) (1U<<((x) % 32))
#define F(x) bit(X86_FEATURE_##x)

void cpuid_whitelist(int is_priv, uint32_t input, uint32_t *eax, uint32_t *ebx,
                     uint32_t *ecx, uint32_t *edx)
{
#if defined(__x86_64__)
    const uint32_t f_lm = F(LM);
#else
    const uint32_t f_lm = 0;
#endif

    /* cpuid 1.edx */
    const uint32_t supported_word0_x86_features =
        F(FPU) | F(VME) | F(DE) | F(PSE) |
        F(TSC) | F(MSR) | F(PAE) | F(MCE) |
        F(CX8) | F(APIC) | 0 /* Reserved */ | F(SEP) |
        F(MTRR) | F(PGE) | F(MCA) | F(CMOV) |
        F(PAT) | F(PSE36) | 0 /* PSN */ | F(CLFLSH) |
        0 /* Reserved, DS, ACPI */ | F(MMX) |
        F(FXSR) | F(XMM) | F(XMM2) | F(SELFSNOOP) |
        0 /* HTT, TM, Reserved, PBE */;

    /* cpuid 0x80000001.edx */
    const uint32_t supported_word1_x86_features =
        F(FPU) | F(VME) | F(DE) | F(PSE) |
        F(TSC) | F(MSR) | F(PAE) | F(MCE) |
        F(CX8) | F(APIC) | 0 /* Reserved */ | F(SYSCALL) |
        F(MTRR) | F(PGE) | F(MCA) | F(CMOV) |
        F(PAT) | F(PSE36) | 0 /* Reserved */ |
        F(NX) | 0 /* Reserved */ | F(MMXEXT) | F(MMX) |
        F(FXSR) | F(FFXSR) | 0 /* GBPAGES, RDTSCP */ |
        0 /* Reserved */ | f_lm | F(3DNOWEXT) | F(3DNOW);

    /* cpuid 1.ecx */
    const uint32_t supported_word4_x86_features =
        F(XMM3) | F(PCLMULQDQ) | 0 /* DTES64, MONITOR */ |
        0 /* DS-CPL */ | F(VMXE) | 0 /* SMXE */ | (is_priv ? F(EST) : 0) |
        0 /* TM2 */ | F(SSSE3) | 0 /* CNXT-ID */ | 0 /* Reserved */ |
        0 /* Reserved */ | F(CX16) | 0 /* xTPR Update, PDCM */ |
        0 /* Reserved, DCA */ | F(XMM4_1) | F(XMM4_2) | 0 /* X2APIC */ |
        F(MOVBE) | F(POPCNT) | 0 /* Reserved*/ | F(AES) | 0 /* XSAVE */ |
        0 /* OSXSAVE, AVX */ | F(F16C) | F(RDRAND) | F(HYPERVISOR);

    /* cpuid 0x80000001.ecx */
    const uint32_t supported_word6_x86_features =
        F(LAHF_LM) | F(CMP_LEGACY) | F(SVM) | 0 /* ExtApicSpace */ |
        F(CR8_LEGACY) | F(ABM) | F(SSE4A) | F(MISALIGNSSE) |
        F(3DNOWPREFETCH) | 0 /* OSVW, IBS */ | F(XOP) |
        0 /* SKINIT, WDT, LWP */ | F(FMA4) | F(TBM);

    /* don't worry about the hypervisor leaves */
    if ( input & 0x40000000 )
        return;

    switch ( input )
    {
    case 0:
        *eax = min(*eax, (uint32_t)0xa);
        break;
    case 1:
        *edx &= supported_word0_x86_features;
        *ecx &= supported_word4_x86_features;
        break;
    case 0x80000000:
        *eax = min(*eax, (uint32_t)0x8000001d);
        break;
    case 0x80000001:
        *edx &= supported_word1_x86_features;
        *ecx &= supported_word6_x86_features;
        break;

    /* see the AMD CPUID specification (pub#25481):
     * Thermal and Power Management
     */
    case 6:
        /* ecx bit 0: EffFreq: effective frequency interface */
        *ecx &= is_priv ? bit(0) : 0;

        /* reserved */
        *eax = *ebx = *edx = 0;
        break;

    /* Advanced Power Management Information */
    case 0x80000007:
        /* edx bit 7: HwPstate: hardware P-state control
         * edx bit 2: VID: Voltage ID control
         * edx bit 1: FID: Frequency ID control
         */
        *edx &= is_priv ? (bit(7) | bit(2) | bit(1)) : 0;

        /* reserved */
        *eax = *ebx = *ecx = 0;
        break;

    /* passthrough */
    case 2:
    case 4:
    case 0x80000002 ... 0x80000006:
    case 0x80000008:
    case 0x8000001a:
    case 0x8000001d:
        break;

    /* zero out everything else */
    default:
        *eax = *ebx = *ecx = *edx = 0;
        break;
    }
}
