/*
 * p2m-ept.c: use the EPT page table as p2m
 * Copyright (c) 2007, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307 USA.
 */

#include <xen/config.h>
#include <xen/domain_page.h>
#include <xen/sched.h>
#include <asm/current.h>
#include <asm/types.h>
#include <asm/domain.h>
#include <asm/hvm/vmx/vmx.h>
#include <xen/iocap.h>
#include <asm/mtrr.h>

#if 1 /* XEN_VERSION == 3 && XEN_SUBVERSION < 2 */

static int ept_set_entry(struct domain *d, unsigned long gfn, mfn_t mfn,
                                int order, p2m_type_t p2mt);
mfn_t ept_get_entry(struct domain *d, unsigned long gfn, p2m_type_t *t);
static mfn_t ept_get_entry_fast(unsigned long gfn, p2m_type_t *t);

static p2m_type_t ept_flags_to_p2m_type(u32 l1e_flags)
{
    if ( l1e_flags & _PAGE_RW )
        return p2m_ram_rw;
    else if ( paging_mode_log_dirty(current->domain) )
        return p2m_ram_logdirty;
    return p2m_invalid;
}

static inline int
compat_ept_set_entry(struct domain *d, unsigned long gfn,
  mfn_t mfn, int order, u32 l1e_flags)
{
    p2m_type_t t = ept_flags_to_p2m_type(l1e_flags);
    if ( t == p2m_ram_rw && mfn_x(mfn) != INVALID_MFN &&
         iomem_access_permitted(d, mfn_x(mfn), mfn_x(mfn)) )
        t = p2m_mmio_direct;

    return ept_set_entry(d, gfn, mfn, order, t);
}

static mfn_t compat_ept_get_entry(struct domain *d, unsigned long gfn)
{
    p2m_type_t dummy;
    return ept_get_entry(d, gfn, &dummy);
}

static mfn_t compat_ept_get_entry_fast(unsigned long gfn)
{
    p2m_type_t dummy;
    return ept_get_entry_fast(gfn, &dummy);
}
#else

#define compat_ept_set_entry ept_set_entry
#define compat_ept_get_entry ept_get_entry
#define compat_ept_get_entry_fast 

#endif

uint8_t epte_get_entry_emt(
    struct domain *d, unsigned long gfn,
    unsigned long mfn, uint8_t *igmt, int direct_mmio)
{
    struct vcpu *v = current;

    *igmt = 0;

    if ( (current->domain != d) && ((v = d->vcpu[0]) == NULL) )
        return MTRR_TYPE_WRBACK;

    if ( !mfn_valid(mfn) )
        return MTRR_TYPE_UNCACHABLE;

    if ( !iommu_enabled )
    {
        *igmt = 1;
        return MTRR_TYPE_WRBACK;
    }

    if ( direct_mmio )
        return MTRR_TYPE_UNCACHABLE;

    if ( iommu_snoop )
    {
        *igmt = 1;
        return MTRR_TYPE_WRBACK;
    }

    return MTRR_TYPE_WRBACK; 
}

static void ept_p2m_type_to_flags(ept_entry_t *entry, p2m_type_t type)
{
    switch(type)
    {
        case p2m_ram_rw:
             entry->r = entry->w = entry->x = 1;
            return;
        case p2m_mmio_direct:
             entry->r = entry->w = entry->x = 1;
            return;
        case p2m_ram_logdirty:
        case p2m_ram_ro:
             entry->r = entry->x = 1;
             entry->w = 0;
            return;
        case p2m_invalid:
        case p2m_mmio_dm:
        default:
            return;
    }
}

#define GUEST_TABLE_NORMAL_PAGE 1
#define GUEST_TABLE_SUPER_PAGE  2
#define GUEST_TABLE_SPLIT_PAGE  3

static struct page_info *ept_alloc_middle_page(struct domain *d)
{
    struct page_info *pg;

    pg = d->arch.p2m.alloc_page(d);
    if ( pg == NULL )
        return NULL;

    pg->count_info = 1;
    pg->u.inuse.type_info = 1 | PGT_validated;
    list_add_tail(&pg->list, &d->arch.p2m.pages);

    return pg;
}

static ept_entry_t ept_set_middle_entry(ept_entry_t *ept_entry,
                                        struct page_info *pg)
{
    ept_entry_t e;

    e.epte = 0;
    e.mfn = page_to_mfn(pg);
    /* last step */
    e.r = e.w = e.x = 1;
    *ept_entry = e;
    return e;
}

static int ept_next_level(struct domain *d, bool_t read_only,
                          ept_entry_t **table, unsigned long *gfn_remainder,
                          u32 shift, int order)
{
    ept_entry_t ept_entry, *next;
    struct page_info *pg;
    u32 index;

    index = *gfn_remainder >> shift;

    ept_entry = (*table)[index];

    if ( !(ept_entry.epte & 0x7) )
    {
        if ( read_only )
            return 0;

        pg = ept_alloc_middle_page(d);
        if ( pg == NULL )
            return 0;
        ept_entry = ept_set_middle_entry((*table) + index, pg);
    }

    if ( !ept_entry.sp_avail )
    {
        *gfn_remainder &= (1UL << shift) - 1;
        next = map_domain_page(ept_entry.mfn);
        unmap_domain_page(*table);
        *table = next;
        return GUEST_TABLE_NORMAL_PAGE;
    }
    else
    {
        if ( order == shift || read_only )
            return GUEST_TABLE_SUPER_PAGE;
        else
            return GUEST_TABLE_SPLIT_PAGE;
    }
}

static int
ept_set_entry(struct domain *d, unsigned long gfn, mfn_t mfn,
                     int order, p2m_type_t p2mt)
{
    ept_entry_t *table = NULL;
    unsigned long gfn_remainder = gfn, offset=0;
    ept_entry_t ept_entry;
    u32 index;
    int i, rv = 0, ret = 0;
    int walk_level = order / EPT_TABLE_ORDER;
    int direct_mmio = (p2mt == p2m_mmio_direct);
    uint8_t igmt = 0; 
    int need_modify_vtd_table = 1;

    /* We only support 4k and 2m pages now */

    BUG_ON(order && order != EPT_TABLE_ORDER);

    if (  order != 0 )
        if ( (gfn & ((1UL << order) - 1)) )
            return 1;

    table = map_domain_page(mfn_x(pagetable_get_mfn(d->arch.phys_table)));

    ASSERT(table != NULL);

    for ( i = EPT_DEFAULT_GAW; i > walk_level; i-- )
    {
        ret = ept_next_level(d, 0, &table,
                    &gfn_remainder, i * EPT_TABLE_ORDER, order);
        if ( !ret )
            goto out;
        else if ( ret != GUEST_TABLE_NORMAL_PAGE )
            break;
    }

    index = gfn_remainder >> ( i ?  (i * EPT_TABLE_ORDER): order);
    walk_level = ( i ? ( i * EPT_TABLE_ORDER) : order) / EPT_TABLE_ORDER;
    offset = (gfn_remainder & ( ((1 << (i*EPT_TABLE_ORDER)) - 1)));

    ept_entry = table[index];

    if ( ret != GUEST_TABLE_SPLIT_PAGE )
    {
        if ( mfn_valid(mfn_x(mfn)) || (p2mt == p2m_mmio_direct) )
        {
            ept_entry.emt = epte_get_entry_emt(d, gfn, mfn_x(mfn),
                                &igmt, direct_mmio);
            ept_entry.igmt = igmt;
            ept_entry.sp_avail = walk_level ? 1 : 0;

            if ( ret == GUEST_TABLE_SUPER_PAGE )
            {
                if ( ept_entry.mfn == (mfn_x(mfn) - offset) )
                    need_modify_vtd_table = 0;
                else
                    ept_entry.mfn = mfn_x(mfn) - offset;

                if ( ept_entry.avail1 == p2m_ram_logdirty &&
                  p2mt == p2m_ram_rw )
                    for ( i = 0; i < (1UL << order); i++ )
                        paging_mark_dirty(d, mfn_x(mfn)-offset+i);
            }
            else
            {
                if ( ept_entry.mfn == mfn_x(mfn) )
                    need_modify_vtd_table = 0;
                else
                    ept_entry.mfn = mfn_x(mfn);
            }


            ept_entry.avail1 = p2mt;
            ept_entry.rsvd = 0;
            ept_entry.avail2 = 0;
            /* last step */
            ept_entry.r = ept_entry.w = ept_entry.x = 1;
            ept_p2m_type_to_flags(&ept_entry, p2mt);
        }
        else
            ept_entry.epte = 0;

        table[index] = ept_entry;
    }
    else
    {
        /* It's super page before, now set one of the 4k pages, so
         * we should split the 2m page to 4k pages now.
         */

        ept_entry_t *split_table = NULL;
        ept_entry_t split_ept_entry;
        unsigned long split_mfn = ept_entry.mfn;
        p2m_type_t split_p2mt = ept_entry.avail1;
        struct page_info *pg;

        /* alloc new page for new ept middle level entry which is
         * before a leaf super entry
         */

        pg = ept_alloc_middle_page(d);
        if ( pg == NULL )
            goto out;

        /* split the super page before to 4k pages */

        split_table = map_domain_page(page_to_mfn(pg));
        offset = gfn & ((1 << EPT_TABLE_ORDER) - 1);

        for ( i = 0; i < 512; i++ )
        {
            split_ept_entry = split_table[i];
            split_ept_entry.emt = epte_get_entry_emt(d,
                                        gfn-offset+i, split_mfn+i,
                                        &igmt, direct_mmio);
            split_ept_entry.igmt = igmt;

            split_ept_entry.sp_avail =  0;

            split_ept_entry.mfn = split_mfn+i;

            split_ept_entry.avail1 = split_p2mt;
            split_ept_entry.rsvd = 0;
            split_ept_entry.avail2 = 0;
            /* last step */
            split_ept_entry.r = split_ept_entry.w = split_ept_entry.x = 1;
            ept_p2m_type_to_flags(&split_ept_entry, split_p2mt);
            split_table[i] = split_ept_entry;
        }

        /* Set the destinated 4k page as normal */
        split_ept_entry = split_table[offset];
        split_ept_entry.emt = epte_get_entry_emt(d, gfn, mfn_x(mfn),
                                                &igmt, direct_mmio);
        split_ept_entry.igmt = igmt;
        if ( split_ept_entry.mfn == mfn_x(mfn) )
            need_modify_vtd_table = 0;
        else
            split_ept_entry.mfn = mfn_x(mfn);

        split_ept_entry.avail1 = p2mt;
        ept_p2m_type_to_flags(&split_ept_entry, p2mt);
        split_table[offset] = split_ept_entry;

        unmap_domain_page(split_table);

        ept_set_middle_entry(table + index, pg);
    }

    /* Track the highest gfn for which we have ever had a valid mapping */
    if ( mfn_valid(mfn_x(mfn))
         && (gfn + (1UL << order) - 1 > d->arch.p2m.max_mapped_pfn) )
        d->arch.p2m.max_mapped_pfn = gfn + (1UL << order) - 1;

    /* Success */
    rv = 1;

 out:
    unmap_domain_page(table);
    ept_sync_domain(d);

    /* support pci pass-through */
    if ( iommu_enabled && is_hvm_domain(d) 
            && need_modify_vtd_table)
    {
        if ( p2mt == p2m_ram_rw )
        {
            if ( order == EPT_TABLE_ORDER )
            {
                for ( i = 0; i < ( 1 << order ); i++ )
                    iommu_map_page(d, gfn-offset+i, mfn_x(mfn)-offset+i);
            }
            else if ( !order )
                iommu_map_page(d, gfn, mfn_x(mfn));
        }
        else
        {
            if ( order == EPT_TABLE_ORDER )
            {
                for ( i = 0; i < ( 1 << order ); i++ )
                    iommu_unmap_page(d, gfn-offset+i);
            }
            else if ( !order )
                iommu_unmap_page(d, gfn);
        }
    }

    return rv;
}

/* Read ept p2m entries */
mfn_t ept_get_entry(struct domain *d, unsigned long gfn, p2m_type_t *t)
{
    ept_entry_t *table =
        map_domain_page(mfn_x(pagetable_get_mfn(d->arch.phys_table)));
    unsigned long gfn_remainder = gfn;
    ept_entry_t ept_entry;
    u32 index;
    int i, ret=0;
    mfn_t mfn = _mfn(INVALID_MFN);

    *t = p2m_mmio_dm;

    /* This pfn is higher than the highest the p2m map currently holds */
    if ( gfn > d->arch.p2m.max_mapped_pfn )
        goto out;

    /* should check if gfn obeys GAW here */

    for ( i = EPT_DEFAULT_GAW; i > 0; i-- )
    {
        ret = ept_next_level(d, 1, &table, &gfn_remainder,
                                        i * EPT_TABLE_ORDER, 0);
        if ( !ret )
            goto out;
        else if ( ret == GUEST_TABLE_SUPER_PAGE )
            break;
    }

    index = gfn_remainder >> ( i * EPT_TABLE_ORDER);
    ept_entry = table[index];

    if ( ept_entry.avail1 != p2m_invalid )
    {
        *t = ept_entry.avail1;
        mfn = _mfn(ept_entry.mfn);
        if ( i )
        {
            /* we may meet super pages, and to split into 4k pages
             * to emulate p2m table
             */
            unsigned long split_mfn = 
              mfn_x(mfn) + (gfn_remainder & ( ((1 << (i*EPT_TABLE_ORDER)) - 1 )));
            mfn = _mfn(split_mfn);
        }
    }

 out:
    unmap_domain_page(table);
    return mfn;
}

static mfn_t ept_get_entry_fast(unsigned long gfn, p2m_type_t *t)
{
    return ept_get_entry(current->domain, gfn, t);
}

/* Walk the whole p2m table, changing any entries of the old type
 * to the new type.  This is used in hardware-assisted paging to
 * quickly enable or diable log-dirty tracking */

static void ept_change_entry_type_global(struct domain *d,
                                         p2m_type_t ot, p2m_type_t nt)
{
    ept_entry_t *l4e, *l3e, *l2e, *l1e;
    int i4, i3, i2, i1;

    if ( pagetable_get_pfn(d->arch.phys_table) == 0 )
        return;

    BUG_ON(EPT_DEFAULT_GAW != 3);

    l4e = map_domain_page(mfn_x(pagetable_get_mfn(d->arch.phys_table)));
    for (i4 = 0; i4 < EPT_PAGETABLE_ENTRIES; i4++ )
    {
        if ( !l4e[i4].epte )
            continue;
        if ( !l4e[i4].sp_avail )
        {
            l3e = map_domain_page(l4e[i4].mfn);
            for ( i3 = 0; i3 < EPT_PAGETABLE_ENTRIES; i3++ )
            {
                if ( !l3e[i3].epte )
                    continue;
                if ( !l3e[i3].sp_avail )
                {
                    l2e = map_domain_page(l3e[i3].mfn);
                    for ( i2 = 0; i2 < EPT_PAGETABLE_ENTRIES; i2++ )
                    {
                        if ( !l2e[i2].epte )
                            continue;
                        if ( !l2e[i2].sp_avail )
                        {
                            l1e = map_domain_page(l2e[i2].mfn);
                            for ( i1  = 0; i1 < EPT_PAGETABLE_ENTRIES; i1++ )
                            {
                                if ( !l1e[i1].epte )
                                    continue;
                                if ( l1e[i1].avail1 != ot )
                                    continue;
                                l1e[i1].avail1 = nt;
                                ept_p2m_type_to_flags(l1e+i1, nt);
                            }
                            unmap_domain_page(l1e);
                        }
                        else
                        {
                            if ( l2e[i2].avail1 != ot )
                                continue;
                            l2e[i2].avail1 = nt;
                            ept_p2m_type_to_flags(l2e+i2, nt);
                        }
                    }
                    unmap_domain_page(l2e);
                }
                else
                {
                    if ( l3e[i3].avail1 != ot )
                        continue;
                    l3e[i3].avail1 = nt;
                    ept_p2m_type_to_flags(l3e+i3, nt);
                }
            }
            unmap_domain_page(l3e);
        }
        else
        {
            if ( l4e[i4].avail1 != ot )
                continue;
            l4e[i4].avail1 = nt;
            ept_p2m_type_to_flags(l4e+i4, nt);
        }
    }
    unmap_domain_page(l4e);

    ept_sync_domain(d);
}

static void __ept_change_entry_type_global(struct domain *d,
                                         u32 l1e_flags)
{
    p2m_type_t nt,ot;

    if ( l1e_flags  == (__PAGE_HYPERVISOR|_PAGE_USER) )
    {
        nt = p2m_ram_rw;
        ot = p2m_ram_logdirty;
    }
    else if ( l1e_flags == (_PAGE_PRESENT|_PAGE_USER) )
    {
        nt = p2m_ram_logdirty;
        ot = p2m_ram_rw;
    }
    else
    {
        nt = ot = p2m_ram_rw;
        BUG();
    }

    ept_change_entry_type_global(d, ot, nt);
}

void ept_p2m_init(struct domain *d)
{
    d->arch.p2m.set_entry = compat_ept_set_entry;
    d->arch.p2m.get_entry = compat_ept_get_entry;
    d->arch.p2m.get_entry_fast = compat_ept_get_entry_fast;
    d->arch.p2m.change_entry_type_global = __ept_change_entry_type_global;
}

/*
 * Local variables:
 * mode: C
 * c-set-style: "BSD"
 * c-basic-offset: 4
 * tab-width: 4
 * indent-tabs-mode: nil
 * End:
 */
