/******************************************************************************
 * include/asm-x86/paging.h
 *
 * physical-to-machine mappings for automatically-translated domains.
 *
 * Copyright (c) 2007 Advanced Micro Devices (Wei Huang)
 * Parts of this code are Copyright (c) 2006 by XenSource Inc.
 * Parts of this code are Copyright (c) 2006 by Michael A Fetterman
 * Parts based on earlier work by Michael A Fetterman, Ian Pratt et al.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _XEN_P2M_H
#define _XEN_P2M_H

#include <asm/paging.h>

#if 1 /* XEN_VERSION == 3 && XEN_SUBVERSION < 2 */

typedef enum {
    p2m_invalid = 0,            /* Nothing mapped here */
    p2m_ram_rw = 1,             /* Normal read/write guest RAM */
    p2m_ram_logdirty = 2,       /* Temporarily read-only for log-dirty */
    p2m_ram_ro = 3,             /* Read-only; writes go to the device model */
    p2m_mmio_dm = 4,            /* Reads and write go to the device model */
    p2m_mmio_direct = 5,        /* Read/write mapping of genuine MMIO area */
} p2m_type_t;

/* We use bitmaps and maks to handle groups of types */
#define p2m_to_mask(_t) (1UL << (_t))

/* RAM types, which map to real machine frames */
#define P2M_RAM_TYPES (p2m_to_mask(p2m_ram_rw)          \
                       | p2m_to_mask(p2m_ram_logdirty)  \
                       | p2m_to_mask(p2m_ram_ro))

/* MMIO types, which don't have to map to anything in the frametable */
#define P2M_MMIO_TYPES (p2m_to_mask(p2m_mmio_dm)        \
                        | p2m_to_mask(p2m_mmio_direct))

/* Read-only types, which must have the _PAGE_RW bit clear in their PTEs */
#define P2M_RO_TYPES (p2m_to_mask(p2m_ram_logdirty)     \
                      | p2m_to_mask(p2m_ram_ro))

/* Useful predicates */
#define p2m_is_ram(_t) (p2m_to_mask(_t) & P2M_RAM_TYPES)
#define p2m_is_mmio(_t) (p2m_to_mask(_t) & P2M_MMIO_TYPES)
#define p2m_is_readonly(_t) (p2m_to_mask(_t) & P2M_RO_TYPES)
#define p2m_is_valid(_t) (p2m_to_mask(_t) & (P2M_RAM_TYPES | P2M_MMIO_TYPES))

/* PTE flags for the various types of p2m entry */
#define P2M_BASE_FLAGS \
        (_PAGE_PRESENT | _PAGE_USER | _PAGE_DIRTY | _PAGE_ACCESSED)

/* Extract the type from the PTE flags that store it */
static inline p2m_type_t p2m_flags_to_type(unsigned long flags)
{
    if ( (flags & _PAGE_RW) && (flags & _PAGE_PCD) )
        return p2m_mmio_direct;
    else if ( flags & _PAGE_RW )
        return p2m_ram_rw;
    else if ( paging_mode_log_dirty(current->domain) )
        return p2m_ram_logdirty;
    else
        return p2m_invalid;
}

static inline unsigned long p2m_type_to_flags(p2m_type_t t)
{
    unsigned long flags = 0;
    switch(t)
    {
    case p2m_ram_rw:
        return flags | P2M_BASE_FLAGS | _PAGE_RW;
    case p2m_mmio_direct:
        return flags | P2M_BASE_FLAGS | _PAGE_RW | _PAGE_PCD;
    case p2m_ram_logdirty:
    case p2m_ram_ro:
    case p2m_mmio_dm:
        return flags | P2M_BASE_FLAGS;
    case p2m_invalid:
    default:
        return flags;
    }
}

#endif

/* The phys_to_machine_mapping is the reversed mapping of MPT for full
 * virtualization.  It is only used by shadow_mode_translate()==true
 * guests, so we steal the address space that would have normally
 * been used by the read-only MPT map.
 */
#define phys_to_machine_mapping ((l1_pgentry_t *)RO_MPT_VIRT_START)


/* Read the current domain's P2M table. */
static inline mfn_t gfn_to_mfn_current(unsigned long gfn)
{
    return current->domain->arch.p2m.get_entry_fast(gfn);
}

/* Read the current domain's p2m table (through the linear mapping). */
static inline mfn_t p2m_gfn_to_mfn_fast(unsigned long gfn)
{
    l1_pgentry_t l1e = l1e_empty();
    l2_pgentry_t l2e = l2e_empty();
    int ret;
    paddr_t addr = ((paddr_t)gfn) << PAGE_SHIFT;
#if CONFIG_PAGING_LEVELS >= 4
    l3_pgentry_t l3e = l3e_empty();
#endif

    if ( gfn > current->domain->arch.p2m.max_mapped_pfn )
        return _mfn(INVALID_MFN);

    /* Don't read off the end of the p2m table */
    ASSERT(gfn < (RO_MPT_VIRT_END - RO_MPT_VIRT_START) / sizeof(l1_pgentry_t));


#if CONFIG_PAGING_LEVELS >= 4
    /* check whether 1GB is available or not */
    ret = __copy_from_user(&l3e,
                           &__linear_l2_table[l2_linear_offset(RO_MPT_VIRT_START) + l3_linear_offset(addr)],
                           sizeof(l3e));
    if ( (ret == 0) && (l3e_get_flags(l3e) & _PAGE_PRESENT) &&
         (l3e_get_flags(l3e) & _PAGE_PSE) )
    {
	return _mfn(l3e_get_pfn(l3e) +
		    l2_table_offset(addr) * L1_PAGETABLE_ENTRIES +
		    l1_table_offset(addr));
    }
#endif

    /* check 2MB entry */
    ret = __copy_from_user(&l2e,
                           &__linear_l1_table[l1_linear_offset(RO_MPT_VIRT_START) + l2_linear_offset(addr)],
                           sizeof(l2e));

    if ( (ret == 0) && (l2e_get_flags(l2e) & _PAGE_PRESENT) &&
         (l2e_get_flags(l2e) & _PAGE_PSE) )
    {
        return _mfn(l2e_get_pfn(l2e) + l1_table_offset(addr));
    }
    else
    {
        ret = __copy_from_user(&l1e,
                               &phys_to_machine_mapping[gfn],
                               sizeof(l1e));
        
        if ( (ret == 0) && (l1e_get_flags(l1e) & _PAGE_PRESENT) )
            return _mfn(l1e_get_pfn(l1e));
    }

    return _mfn(INVALID_MFN);
}

/* Read another domain's P2M table, mapping pages as we go */
static inline
mfn_t gfn_to_mfn_foreign(struct domain *d, unsigned long gpfn)
{
    return d->arch.p2m.get_entry(d, gpfn);
}

/* General conversion function from gfn to mfn */
static inline mfn_t gfn_to_mfn(struct domain *d, unsigned long gfn)
{
    if ( !paging_mode_translate(d) )
        return _mfn(gfn);
    if ( likely(current->domain == d) )
        return gfn_to_mfn_current(gfn);
    else 
        return gfn_to_mfn_foreign(d, gfn);
}

/* General conversion function from mfn to gfn */
static inline unsigned long mfn_to_gfn(struct domain *d, mfn_t mfn)
{
    if ( paging_mode_translate(d) )
        return get_gpfn_from_mfn(mfn_x(mfn));
    else
        return mfn_x(mfn);
}

/* Compatibility function for HVM code */
static inline unsigned long get_mfn_from_gpfn(unsigned long pfn)
{
    return mfn_x(gfn_to_mfn_current(pfn));
}

/* Is this guest address an mmio one? (i.e. not defined in p2m map) */
static inline int mmio_space(paddr_t gpa)
{
    unsigned long gfn = gpa >> PAGE_SHIFT;
    return !mfn_valid(mfn_x(gfn_to_mfn_current(gfn)));
}

/* Translate the frame number held in an l1e from guest to machine */
static inline l1_pgentry_t
gl1e_to_ml1e(struct domain *d, l1_pgentry_t l1e)
{
    if ( unlikely(paging_mode_translate(d)) )
        l1e = l1e_from_pfn(gmfn_to_mfn(d, l1e_get_pfn(l1e)),
                           l1e_get_flags(l1e));
    return l1e;
}


/* Set mmio addresses in the p2m table (for pass-through) */
int set_mmio_p2m_entry(struct domain *d, unsigned long gfn, mfn_t mfn);
int clear_mmio_p2m_entry(struct domain *d, unsigned long gfn);

/* Init the datastructures for later use by the p2m code */
void p2m_init(struct domain *d);

/* Allocate a new p2m table for a domain. 
 *
 * The alloc_page and free_page functions will be used to get memory to
 * build the p2m, and to release it again at the end of day. 
 *
 * Returns 0 for success or -errno. */
int p2m_alloc_table(struct domain *d,
                    struct page_info * (*alloc_page)(struct domain *d),
                    void (*free_page)(struct domain *d, struct page_info *pg));

/* Return all the p2m resources to Xen. */
void p2m_teardown(struct domain *d);

/* Add a page to a domain's p2m table */
int guest_physmap_add_page(struct domain *d, unsigned long gfn,
                            unsigned long mfn, int order);

/* Remove a page from a domain's p2m table */
void guest_physmap_remove_page(struct domain *d, unsigned long gfn,
                               unsigned long mfn, int order);

/* set P2M table l1e flags */
void p2m_set_flags_global(struct domain *d, u32 l1e_flags);

void p2m_change_entry_type_global(struct domain *d, u32 l1e_flags);

/* set P2M table l1e flags for a gpa */
int p2m_set_flags(struct domain *d, paddr_t gpa, u32 l1e_flags);

#endif /* _XEN_P2M_H */

/*
 * Local variables:
 * mode: C
 * c-set-style: "BSD"
 * c-basic-offset: 4
 * indent-tabs-mode: nil
 * End:
 */
